#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2017-2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import re
from pathlib import Path
from collections import defaultdict

base_path = Path(__file__).resolve().parent.parent
if __name__ == '__main__':
    sys.path.insert(0, str(base_path))

import buildlib.utils

sep = ' \\\n    '


def read_files_meta(filename, srcdir=None):
    pwd = os.getcwd()
    os.chdir(srcdir or base_path)
    try:
        files_meta = buildlib.utils.FileListsSetup(filename)
        files_meta.parse()
    finally:
        os.chdir(pwd)
    if files_meta.errors:
        print(f'{files_meta.errors} errors in "{filename}"')
        sys.exit(1)
    return files_meta
    
    
def cmd_modify(source, target, modulename, *variants):
    source = Path(source)
    assert source.parts[0] == '..', source
    files_meta = read_files_meta('files-meta')
    attrs = files_meta.allfiles[Path(*source.parts[1:])]
    variantflags = buildlib.utils.filter_variant_by_groups(variants, attrs.variantgroup)
    vflags = tuple(v for v in variants if v not in buildlib.utils.variant_names)
    buildlib.utils.modify_variant_file(files_meta, source, Path(target), modulename, variantflags, vflags)
    
    
def find_ext_sources(files_meta, variants):
    sources = defaultdict(list)
    depends = defaultdict(list)
    
    for filename, attrs in files_meta.ext:
        if any((v not in variants) for v in attrs.variantonly):
            continue
        filename = Path('..', filename)
        name = buildlib.utils.get_modulename(filename)
        variantflags = buildlib.utils.filter_variant_by_groups(variants, attrs.variantgroup)
        modulename = buildlib.utils.get_modulevariantname(name, variantflags)
        sources[filename.suffix].append((modulename, filename, filename.stem, attrs))
        
        for filename, attrs in files_meta.allfiles.items():
            if 'depends' not in attrs or name not in attrs.depends:
                continue
            if any(f not in variants for f in attrs.variantonly):
                continue
            filename = Path('..', filename)
            depends[filename.suffix].append((modulename, filename, filename.stem, attrs))
    return sources, depends
    
def rules_freeze(files_meta, sources):
    text = '#### rules: freeze scripts ####\n'
    command = '$$python_interpreter ../buildlib/gentools.py'
    assert len(list(files_meta.script)) == 1
    for filename, unused_attrs in files_meta.script:
        in_file = Path('..', filename)
        cmodule = 'B_'+in_file.stem
        out_file = Path(cmodule+'.c')
        text += (f'rule_freeze_{cmodule}.target = {out_file}\n'
                 f'rule_freeze_{cmodule}.commands = {command} freeze-module{sep}{in_file}{sep}{out_file} __main__\n'
                 f'rule_freeze_{cmodule}.depends = {in_file}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_freeze_{cmodule}\n'
                 f'FROZEN_MODULES += {out_file}\n'
                 '\n')
        sources['.c/cpp'].append((None, out_file, cmodule, None))
    text += '#### rules: freeze application python modules ####\n'
    for filename, unused_attrs in files_meta.py:
        pymodule = str(filename.with_suffix(''))
        cmodule = 'M_'+pymodule.replace('/', '__')
        in_file = Path('..', filename)
        out_file = Path(cmodule+'.c')
        text += (f'rule_freeze_{cmodule}.target = {out_file}\n'
                 f'rule_freeze_{cmodule}.commands = {command} freeze-module{sep}{in_file}{sep}{out_file} {pymodule}\n'
                 f'rule_freeze_{cmodule}.depends = {in_file}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_freeze_{cmodule}\n'
                 f'FROZEN_MODULES += {out_file}\n'
                 '\n')
        sources['.c/cpp'].append((None, out_file, cmodule, None))
    return text
    
def _find_extension_module(filename):
    import importlib.util
    def guessmodule(module):
        yield module
        if module.endswith('module'):
            module = module.rsplit('module',1)[0]
            yield module
        if module[0] != '_':
            module = '_'+module
            yield module
            
    for module in guessmodule(filename.stem):
        m = importlib.util.find_spec(module)
        if m is not None and (m.origin == 'built-in' or m.loader.__class__.__name__ in ['ExtensionFileLoader']):
            return module
    return None
    
def rules_freeze_python(files_meta):
    text = '#### rules: freeze stdlib python modules ####\n'
    command = '$$python_interpreter $$PWD/gentools.py'
    for in_file in sorted(files_meta.includes):
        if in_file.parts[0] != 'Lib':
            continue
        pymodule = str(Path(*in_file.parts[1:]).with_suffix(''))
        cmodule = 'S_' + pymodule.replace('/', '__')
        out_file = Path(cmodule+'.c')
        text += (f'rule_freeze_{cmodule}.target = {out_file}\n'
                 f'rule_freeze_{cmodule}.commands = {command} freeze-module{sep}$$PWD/{in_file}{sep}{out_file} {pymodule}\n'
                 f'rule_freeze_{cmodule}.depends = $$PWD/{in_file}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_freeze_{cmodule}\n'
                 f'FROZEN_MODULES += {out_file}\n'
                 f'GENERATED_SOURCES += {out_file}\n'
                 '\n')
    text += '#### stdlib extension modules ####\nSTDLIB_EXTENSION_MODULES = \\\n'
    for in_file in sorted(files_meta.includes):
        if in_file.parts[0] != 'Modules':
            continue
        module = _find_extension_module(in_file)
        if module is not None:
            text += f'    {module} \\\n'
    text += f'\n'
    return text
    
def rules_copy_depends(depends):
    text = '#### rules: copy depends .h .ui .cpp ####\n'
    for filename, key in buildlib.utils.uniq((f,k) for s in ('.h','.ui', '.cpp') for m,f,k,a in depends[s]):
        suffix_rule = filename.suffix.replace('.', '')
        text += (f'rule_modify_{key}_{suffix_rule}.target = {filename.name}\n'
                 f'rule_modify_{key}_{suffix_rule}.commands = cp{sep}{filename} {filename.name}\n'
                 f'rule_modify_{key}_{suffix_rule}.depends = {filename}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_modify_{key}_{suffix_rule}\n'
                 '\n')
    depends['.h'] = [(m, Path(f.name), k ,a) for m,f,k,a in depends['.h']]
    depends['.ui'] = [(m, Path(f.name), k ,a) for m,f,k,a in depends['.ui']]
    depends['.cpp'] = [(m, Path(f.name), k ,a) for m,f,k,a in depends['.cpp']]
    return text
    
def rules_modify_depends(variants, depends):
    text = '#### rules: modify depends .pxd ####\n'
    command = '../buildlib/gentools.py'
    variants = ' '.join(variants)
    filenames = []
    for variantname, filename, key, attrs in depends['.pxd']:
        if filename in filenames:
            continue
        filenames.append(filename)
        text += (f'rule_modify_pxd_{key}.target = {filename.name}\n'
                 f'rule_modify_pxd_{key}.commands = {command} modify{sep}{filename} {filename.name} {variantname} {variants}\n'
                 f'rule_modify_pxd_{key}.depends = {filename}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_modify_pxd_{key}\n'
                 f'QMAKE_CLEAN += {filename.name}\n'
                 '\n')
    depends['.pxd'] = [(m, Path(f.name), k, a) for m,f,k,a in depends['.pxd']]
    return text
    
def rules_modify_sources(variants, sources, depends):
    text = '#### rules: modify sources .py .pyx ####\n'
    command = '../buildlib/gentools.py'
    targets = defaultdict(list)
    variants = ' '.join(variants)
    modules = []
    for suffix in '.py', '.pyx':
        for name, filename, key, attrs in sources[suffix]:
            assert suffix == filename.suffix
            out_file = Path(name+filename.suffix)
            text += (f'rule_modify_src{name}.target = {out_file}\n'
                     f'rule_modify_src{name}.commands = {command} modify{sep}{filename}{sep}{out_file} {name} {variants}\n'
                     f'rule_modify_src{name}.depends = {filename}\n'
                     f'QMAKE_EXTRA_TARGETS += rule_modify_src{name}\n'
                     f'QMAKE_CLEAN += {out_file}\n'
                     '\n')
            modules.append(out_file.stem)
            targets[suffix].append((name, out_file, key, attrs))
            if suffix in depends:
                for i, (dname, dfilename, dkey, dattrs) in enumerate(depends[suffix]):
                    if dfilename == filename:
                        depends[suffix][i] = (dname, out_file, dkey, dattrs)
    text += '#### list extension module names ####\n'
    text += f'EXTENSION_MODULES ={sep}' + f'{sep}'.join(modules) + '\n\n'
    sources.update(targets)
    return text
    
def rules_py2pyx(sources, depends):
    text = '#### rules: py2pyx ####\n'
    def px_files():
        if '.py' in sources:
            for name, filename, key, attrs in sources['.py']:
                yield name, filename, key, attrs, True
        if '.cpp' in depends:
            for name, filename, key, attrs in depends['.cpp']:
                yield name, filename, key, attrs, False
    for name, filename, key, attrs, issource in list(px_files()):
        if 'py2pyx' not in attrs:
            continue
        out_file = filename.with_suffix('.pyx' if issource else '.pyx_unused')
        out_base = filename.with_suffix('')
        lang_ext = '.cpp' if 'c++' in attrs.language else '.c'
        command = '../buildlib/py2pyx.py'
        out_extra = sep.join(f'{out_base}{suffix}' for suffix in attrs.py2pyx)
        text += (f'rule_py2pyx{out_file.stem}_pyx.target = {out_file}\n'
                 f'rule_py2pyx{out_file.stem}_pyx.commands = {command}{sep}{filename}{sep}{out_file}{sep}{out_extra}\n'
                 f'rule_py2pyx{out_file.stem}_pyx.depends = {filename}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_py2pyx{out_file.stem}_pyx\n'
                 f'QMAKE_CLEAN += {out_file}\n'
                 '\n')
        if issource:
            sources['.pyx'].append((name, out_file, key, attrs))
        for suffix in attrs.py2pyx:
            rule_suffix = suffix.replace('.', '_')
            out_extra = Path(f'{out_base}{suffix}')
            text += (f'rule_py2pyx{out_file.stem}{rule_suffix}.target = {out_extra}\n'
                     f'rule_py2pyx{out_file.stem}{rule_suffix}.commands = test -e {out_extra}\n'
                     f'rule_py2pyx{out_file.stem}{rule_suffix}.depends = {out_file}\n'
                     f'QMAKE_EXTRA_TARGETS += rule_py2pyx{out_file.stem}{rule_suffix}\n'
                     f'QMAKE_CLEAN += {out_extra}\n'
                     '\n')
            depends[out_extra.suffix].append((name, out_extra, key, attrs))
    del sources['.py']
    depends['.pxd'] += [(m, f.with_suffix('.pxd'), k, a) for m,f,k,a in depends['.py']]
    del depends['.py']
    return text
        
def rules_cython(sources, depends):
    text = '#### rules: cython ####\n'
    for name, filename, key, attrs in sources['.pyx']:
        lang_flag, lang_ext = ('--cplus','.cpp') if 'c++' in attrs.language else ('','.c')
        out_file = filename.with_suffix(lang_ext)
        out_base = filename.with_suffix('')
        text += (f'rule_cython{name}.target = {out_file}\n'
                 f'rule_cython{name}.commands = cython3 -3 {lang_flag}{sep}{filename}{sep}-o {out_file}\n'
                 f'rule_cython{name}.depends = {filename}\n')
        for dname, dfilename, unused_dkey, dattrs in depends['.pxd']:
            if dname == name:
                text +=  f'rule_cython{name}.depends += {dfilename.name}\n'
        text +=  f'QMAKE_EXTRA_TARGETS += rule_cython{name}\n\n'
        sources['.c/cpp'].append((name, out_file, key, attrs))
    del sources['.pyx']
    del depends['.pxd']
    return text
    
def rules_moc(depends):
    text = '#### rules: moc ####\n'
    for name, filename, key, attrs in depends['.h']:
        if not filename.stem.endswith('_moc'):
            continue
        out_file = filename.with_suffix('.cpp')
        text += (f'rule_moc{key}.target = {out_file}\n'
                 f'rule_moc{key}.commands = $$[QT_INSTALL_BINS]/moc{sep}'
                                f'{filename}{sep}-o {out_file}\n'
                 f'rule_moc{key}.depends ={sep}{filename}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_moc{key}\n'
                 '\n')
        depends['.cpp'].append((name, out_file, key, attrs))
    return text
    
def rules_uic(depends):
    text = '#### rules: uic ####\n'
    for name, filename, key, attrs in depends['.ui']:
        out_file = Path(filename.name).with_suffix('.h')
        text += (f'rule_ui{key}.target = {out_file}\n'
                 f'rule_ui{key}.commands = $$[QT_INSTALL_BINS]/uic --translate gettext_translate{sep}'
                                f'{filename}{sep}-o {out_file}\n'
                 f'rule_ui{key}.depends ={sep}{filename}\n'
                 f'QMAKE_EXTRA_TARGETS += rule_ui{key}\n'
                 '\n')
        depends['.h'].append((name, out_file, key, attrs))
    del depends['.ui']
    return text
    
def list_cfiles(sources, depends):
    text = '#### list sources .c/cpp ####\n'
    for filelist in sources['.c/cpp'], depends['.cpp']:
        for name, filename, unused_key, attrs in filelist:
            text += f'GENERATED_SOURCES += {filename}\n'
    text += '\n#### list depends .h ####\n'
    for name, filename, unused_key, attrs in depends['.h']:
        text += f'PRE_TARGETDEPS += {filename}\n'
    text += '\n'
    del sources['.c/cpp']
    del depends['.cpp']
    del depends['.h']
    return text
    
def rules_data(files_meta):
    text = '#### data ####\n'
    for d, (paths, flags) in files_meta.group_data().items():
        d = Path(d)
        paths = sep.join(f'"../{p}"' for p in paths)
        var = '_'.join(d.parts)
        flags = ''.join(f+':' for f in flags if f in buildlib.utils.variant_names)
        text += f'data_{var}.path = /assets/data/{d}\n'
        text += f'data_{var}.files ={sep}{paths}\n'
        text += f'{flags}INSTALLS += data_{var}\n'
        text += '\n'
    return text
    
def cmd_pybik_files_pri(*args):
    args = iter(args)
    variants = []
    for a in args:
        variants.append(a)
        assert a != '--'
    files_meta = read_files_meta('files-meta')
        
    text = f'# generated by {sys.argv}\n\n'
    sources, depends = find_ext_sources(files_meta, variants)
    text += rules_freeze(files_meta, sources)
    text += rules_copy_depends(depends)
    text += rules_modify_depends(variants, depends)
    text += rules_modify_sources(variants, sources, depends)
    text += rules_py2pyx(sources, depends)
    text += rules_cython(sources, depends)
    text += rules_moc(depends)
    text += rules_uic(depends)
    text += list_cfiles(sources, depends)
    text += rules_data(files_meta)
    
    path = Path('pybik-files.pri')
    try:
        old_text = path.read_text(encoding='utf-8')
    except FileNotFoundError:
        old_text = None
    try:
        if sources or depends:
            print('unused source files:')
            for d,n in (depends,'depends:'), (sources,'sources:'):
                print('  ----', n)
                for suffix, l in d.items():
                    print(f'    {suffix}:')
                    for values in l:
                        print('     ', str(values))
            sys.exit(1)
    finally:
        if text != old_text:
            path.write_text(text, encoding='utf-8')
        
def cmd_freeze_module(source, target, module):
    data = Path(source).read_text(encoding='utf-8')
    data = compile(data, source, 'exec')
    import marshal
    data = marshal.dumps(data)
    data = ''.join(('%d,' if i%16 else '\n    %d,')%c for i,c in enumerate(data))
    size = len(data)
    
    filename = Path(module)
    if filename.name == '__init__':
        module = str(filename.parent)
        size = -size
    pname = module.replace('/', '.')
    cname = 'M_' + module.replace('/', '__')
    
    template = (
        '/*\n'
        'pname: {0}\n'
        'cname: {1}\n'
        'size: {2}\n'
        '*/\n'
        'unsigned char {1}[] = {{{3}\n'
        '}};\n\n'
    )
    
    data = template.format(pname, cname, size, data)
    Path(target).write_text(data, encoding='utf-8')
    
def cmd_freeze_modules(file_frozen, *module_files):
    modules = []
    module_files = iter(module_files)
    for filename in module_files:
        if filename == '--':
            break
        filename = Path(filename)
        data = filename.read_text()
        pname = re.search(r'^pname: (.*)$', data, re.M).group(1)
        cname = re.search(r'^cname: (.*)$', data, re.M).group(1)
        size = re.search(r'^size: (.*)$', data, re.M).group(1)
        modules.append((pname, cname, size))
    for name in module_files:
        modules.append(name)
        
    with open(file_frozen, 'wt', encoding='utf-8') as fout:
        for m in modules:
            print(repr(m), file=fout)
            
def cmd_freeze(target, *frozen_files):
    from ast import literal_eval
    text1 = text2 = text3 = text4 = ''
    for filename in frozen_files:
        text1 += f'\n// {filename}\n'
        text2 += f'\n    // {filename}\n'
        text3 += f'\n// {filename}\n'
        text4 += f'\n    // {filename}\n'
        with open(filename, 'rt') as fin:
            for line in fin:
                line = literal_eval(line)
                if type(line) is tuple:
                    pmodule, cmodule, size = line
                    text1 += f'extern unsigned char {cmodule}[];\n'
                    text2 += f'    {{"{pmodule}", {cmodule}, {size}}},\n'
                else:
                    assert type(line) is str
                    text3 += f'extern PyObject *PyInit_{line}(void);\n'
                    text4 += f'    {{"{line}", PyInit_{line}}},\n'
    with open(target, 'wt', encoding='utf-8') as fout:
        print('\n#include "Python.h"\n'
              f'{text1}\n'
              'struct _frozen _PyImport_FrozenModules[] = {'
              f'{text2}\n'
              '    {0, 0, 0}\n};\n'
              f'{text3}\n'
              'struct _inittab _PyImport_InittabModules[] = {'
              f'{text4}\n'
              '    {0, 0}\n};\n\n'
              , file=fout)
        
def cmd_python_files_pri():
    files_meta = read_files_meta('freezer/python-files-meta', '.')
    
    text = f'# generated by {sys.argv}\n\n'
    text += rules_freeze_python(files_meta)
    
    text += '#### list sources .c/cpp ####\n'
    for filename in files_meta.includes:
        if not filename or filename.parts[0] == 'Lib':
            continue
        text += f'SOURCES += {filename}\n'
    text += '\n'
    
    path = Path('python-files.pri')
    try:
        old_text = path.read_text(encoding='utf-8')
    except FileNotFoundError:
        old_text = None
    if text != old_text:
        path.write_text(text, encoding='utf-8')
        
def cmd_pyconfig_h(source_dir, scope, target_dir):
    defines_file = Path(source_dir, 'pyconfig-defines')
    source_file = Path(target_dir, 'pyconfig.h.orig')
    target_file = Path(target_dir,'pyconfig.h')
    
    print(f'generate {target_file.name}: {source_file} -> {target_file}')
    
    def get_defines():
        from ast import literal_eval
        for line in defines_file.read_text('utf-8').splitlines():
            if not line.startswith('#'):
                k, v = line.strip().split(' ', 1)
                *scopes, k = k.split(':')
                if not scopes or scope in scopes:
                    yield k, ('new' in scopes, literal_eval(v))
    defines = dict(get_defines())
    used = set()
    
    errors = 0
    changes = 0
    slines = Path(source_file).read_text('utf-8').splitlines()
    tlines = []
    
    def append_new_defines():
        nonlocal errors, changes
        for name, (isnew, value) in defines.items():
            if isnew:
                used.add(name)
                if value is None:
                    print(f'error: new name, but value is None: {name!r}')
                    errors += 1
                    continue
                else:
                    tline = f'#define {name} {value}'
                    tlines.append(tline)
                    changes += 1
                    
    for n, line in enumerate(slines):
        if line == '#endif /*Py_PYCONFIG_H*/':
            append_new_defines()
        match = re.match(r'^((/\*)? *# *(undef|define) +([A-Z0-9_]+) *[a-zA-Z0-9]*(\*/)?)$', line)
        name = match and match[4]
        if name not in defines:
            tlines.append(line)
            continue
        if (match[2] is None) != (match[5] is None):
            print(f'error in {n}: unexpected comment symbols: {line!r}')
            errors += 1
            tlines.append(line)
            continue
        isnew, value = defines[name]
        if isnew:
            print(f'error in {n}: match, but marked as new: {line!r}')
            errors += 1
            tlines.append(line)
            continue
        used.add(name)
        if value is None:
            if match[2]:
                print(f'error in {n}: already a comment: {line!r}')
                #errors += 1
                tlines.append(line)
                continue
            else:
                tlines.append(f'/*{line}*/')
            changes += 1
        else:
            tline = f'#define {name} {value}'
            if tline == line:
                print(f'error ({n}): unchanged match: {line!r}')
                errors += 1
                tlines.append(line)
                continue
            tlines.append(tline)
            changes += 1
    target_file.write_text('\n'.join(tlines), encoding='utf-8')
    unused = sorted(k for k in defines if k not in used)
    if unused:
        print(f'error: {len(unused)} unused entries in {defines_file}:')
        unused = ' '.join(unused)
        print(f'    {unused}')
        errors += 1
    print(f'  {changes} changed lines in {source_file}')
    if errors:
        print(f'  {errors} errors for {source_file}')
        sys.exit(1)
        
accepted_commands = ('modify pybik-files.pri freeze-module freeze-modules freeze '
                     'python-files.pri pyconfig.h').split()

def usage():
    print('usage:')
    prog = os.path.basename(__file__)
    print(' ', prog, '-h|--help')
    for cmd in accepted_commands:
        print(' ', prog, cmd)
    sys.exit(1)
    
def main():
    args = sys.argv[1:]
    if len(args) < 1:
        usage()
    cmd, *args = args
    if cmd in ['-h', '--help']:
        usage()
    elif cmd in accepted_commands:
        globals()['cmd_'+cmd.replace('-','_').replace('.','_')](*args)
    else:
        usage()
        
if __name__ == '__main__':
    main()
    

