# -*- coding: utf-8 -*-

#  Copyright © 2013-2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
from glob import iglob
from fnmatch import fnmatch
from contextlib import suppress
from pathlib import Path


def update_linguas(po_files):
    import polib
    min_percent_translated = 20
    cnt_skipped = 0
    with open('po/LINGUAS', 'wt', encoding='utf-8') as linguas:
        po_files = sorted(po_files)
        for po_file, unused_attrs in po_files:
            percent_translated = polib.pofile(po_file).percent_translated()
            if percent_translated < min_percent_translated:
                print('skipping po file {!r} ({}%<{}% translated)'.format(
                            str(po_file), percent_translated, min_percent_translated),
                    file=sys.stderr)
                cnt_skipped += 1
            else:
                print(po_file.stem, file=linguas)
        print('skipped {}/{} po files'.format(cnt_skipped, len(po_files)), file=sys.stderr)
                
def get_linguas():
    with open('po/LINGUAS', 'rt', encoding='utf-8') as linguas:
        return linguas.read().splitlines()
        
def generate_lang_translations():
    from pybiklib.dialogs import get_langname_from_code_func
    langname_from_code = get_langname_from_code_func()
    names = ((code, langname_from_code(code)) for code in get_linguas())
    names = {code: name for code, name in names if name is not None}
    print(names)
    
def generate_all_lang_translations(file=sys.stdout):
    import icu  # to be sure get_langname_from_code_func does not return a fallback function
    from subprocess import Popen, PIPE
    from ast import literal_eval
    from pprint import pformat
    
    processes = []
    env = os.environ.copy()
    for lang in get_linguas():
        env['LC_MESSAGES'] = lang+'.UTF-8'
        script = 'import buildlib.utils as m; m.generate_lang_translations()'
        print(lang, end=' ', flush=True)
        processes.append((lang, Popen(['python3', '-c', script], env=env, stdout=PIPE)))
    print()
    
    langs = {}
    for lang, process in processes:
        print(lang, end=' ', flush=True)
        outstr, unused_err = process.communicate()
        outstr = outstr.strip().decode('utf-8')
        try:
            langs[lang] = literal_eval(outstr)
        except SyntaxError:
            print('code for lang %r:' % lang, repr(outstr))
            raise
    print()
    
    print('# -*- coding: utf-8 -*-', file=file)
    print('# generated file\n', file=file)
    print('languages = (', file=file)
    print(pformat(langs), file=file)
    print(')\n', file=file)
    
    
def modify_file(source, target, subs, count=1):
    from shutil import copymode, copystat
    import re
    print(f'modifying {source} -> {target}')
    text = source.read_text(encoding='utf-8')
    try:
        cmptext = target.read_text(encoding='utf-8')
    except FileNotFoundError as e:
        cmptext = None
    for pattern, repl in subs:
        text, n = re.subn(pattern, repl, text, count=count, flags=re.MULTILINE)
        print(f'    {n}x replaced {pattern!r} with {repl!r}')
    if text == cmptext:
        copystat(source, target)
        return False
    tmpfilename = target.with_name(target.name + '.tmp')
    tmpfilename.write_text(text, encoding='utf-8')
    copymode(source, tmpfilename)
    tmpfilename.rename(target)
    return True
        
    
def uniq(items):
    result = []
    for v in items:
        if v not in result:
            result.append(v)
    return result
    
variants = (
    {'id': 'qtw',  'name': 'QtWidgets (default)',  'group': 'qtvariant',  'type': 'enum'},
    {'id': 'qtq',  'name': 'QtQuick/QML',          'group': 'qtvariant',  'type': 'enum'},
    {'id': 'os',   'name': 'offscreen',            'group': 'offscreen',  'type': 'bool'},
    {'id': 'all',  'name': 'all variants',                                'type': None},
)
vflags = (
    {'id': 'android',   'name': 'Android',            'group': 'android',    'type': 'bool'},
    {'id': 'frozenqtq', 'name': 'frozen QtQuick/QML', 'group': 'frozenqtq',  'type': 'bool'},
)
variant_names = [v['id'] for v in variants if 'group' in v]
variant_groups = uniq(v['group'] for v in variants if 'group' in v)
variants_by_id = {v['id']: v for v in variants}
variants_and_vflags_by_id = {v['id']: v for v in variants+vflags}
    
def variant_string():
    return ', '.join('%s (%s)'%(v['id'],v['name']) for v in variants)
    
def add_defaults(variant_ids):
    default_group = None
    default_variant = None
    result = []
    for v in variants:
        if v['type'] == None:
            continue
        if default_group != v['group']:
            if default_variant is not None:
                result.append(default_variant)
            default_group = v['group']
            default_variant = v['id'] if v['type'] == 'enum' else None
        if v['id'] in variant_ids or 'all' in variant_ids:
            result.append(v['id'])
            default_variant = None
    assert default_variant is None
    return result
    
def variant_aliases():
    groups = []
    aliases = {'': ''}
    for v in variants:
        if v['type'] is None or v['group'] in groups:
            continue
        groups.append(v['group'])
        if v['type'] == 'enum':
            for a, va in list(aliases.items()):
                aliases[a] = va + v['id']
                for ve in variants:
                    if ve['type'] == 'enum' and ve['group'] == v['group']:
                        aliases[a+ve['id']] = va+ve['id']
        elif v['type'] == 'bool':
            for a, va in list(aliases.items()):
                aliases[a+v['id']] = va+v['id']
    return [(k,v) for k,v in aliases.items() if k != v]
    
def all_variants():
    def group_variants():
        grouped_variants = {}
        group = None
        values = None
        for v in variants:
            if v['type'] == None:
                continue
            if group != v['group']:
                group = v['group']
                values = []
                yield values
            if v['type'] == 'bool':
                values.append(None)
            values.append(v['id'])
    result = [()]
    for variantgroup in list(group_variants()):
        result = [(v if f is None else v+(f,)) for v in result for f in variantgroup]
    return result
        
        
is_variant_in_groups = lambda v, g: variants_and_vflags_by_id[v].get('group') in g
filter_variant_by_groups = lambda vf, groups: tuple(f for f in vf if is_variant_in_groups(f, groups))
get_modulename = lambda filename: '_'+filename.stem.lstrip('_').rsplit('_',1)[0]
get_fullmodulename = lambda filename: str(filename.with_name(get_modulename(filename)))
get_modulevariantname = lambda modulename, variantflags: modulename + '_' + ''.join(variantflags)
def get_variantfilename(dirname, filename, variantflags):
    modulename = get_modulename(filename)
    modulevariantname = get_modulevariantname(modulename, variantflags)
    return Path(dirname, filename.parent, modulevariantname + filename.suffix)
    
def modify_variant_file(files_meta, source, target, modulevariantname, variantflags, vflags=None):
    modulesets = files_meta.get_modulesets()
    if vflags is None:  vflags = ()
    variantstr = ''.join(variantflags)
    print(f'changing to variant {variantstr!r} vflags={" ".join(vflags)!r}: {source} -> {target}')
    
    subs = []
    for mvn in modulesets[modulevariantname]:
        modulename = get_modulename(Path(mvn))
        subs.append((fr'\[\[{modulename}_VARIANT\]\]', mvn))
    for v in variantflags+vflags:
        vgroup = variants_and_vflags_by_id[v]['group'].upper()
        subs.append((r'\[\[%s\]\]' % vgroup, v))
        if variants_and_vflags_by_id[v]['type'] == 'bool':
            subs.append((r'\b(DEF %s = )False' % vgroup, r'\1True'))
    def depends_lines():
        yield '{'
        for vs in all_variants():
            joinkey = repr(''.join(vs)) + ':'
            joinvalue = modulesets[vs]
            yield '  {:14} {!r},'.format(joinkey, joinvalue)
        yield '}'
    def aliases_lines():
        yield '{'
        for avs, vs in variant_aliases():
            yield '  {:12} {!r},'.format(repr(avs) + ':', vs)
        yield '}'
    subs += [   (r'\[\[VARIANT\]\]', variantstr),
                (r'\[\[DEPENDS\]\]', lambda m: '\n  '.join(depends_lines())),
                (r'\[\[ALIASES\]\]', lambda m: '\n  '.join(aliases_lines())),
            ]
    modify_file(source, target, subs, count=0)
    
def sync_shader(path, *, modify):
    assert path.suffix in ('.vert', '.frag'), path.suffix
    isvert = path.suffix == '.vert'
    subdata = {
        #'gl3': [
        #    (r' *//gl3.$', ''),
        #],
        'es2': [
            (r'^#version .*$', '#define FP highp'),
            (r'^.*//gl3.$', ''),
            (r'//es2:', ''),
            (r'^in ', ('attribute ' if isvert else 'varying ')),
            (r'^out ', ('varying ' if isvert else '//out ')),
            (r'\b(vec[234] +|float +|sampler2D +)\b', r'\1' if isvert else r'FP \1'),
            ('fragColor', 'gl_FragColor'),
        ],
    }
    copy_modified = lambda src, dst: modify_file(src, dst, subs, count=0)
    from shutil import copy
    for subdir, subs in subdata.items():
        (path.parent/subdir).mkdir(exist_ok=True)
        if modify:
            yield copy_modified, path, path.parent/subdir/path.name
        else:
            yield copy, path, path.parent/subdir/path.name
    
def test_shader_interface(path, interface):
    res = []
    error = False
    lines = Path(path).read_text(encoding='utf-8').splitlines()
    stage = 0
    stage_prefs = ('#', None, 'in ', 'out ', 'uniform ')
    for i,line in enumerate(lines):
        if line.startswith('//'):
            continue
        if not line.strip():
            stage += 1
            continue
        line = line.split('//',1)[0].rstrip()
        try:
            pref = stage_prefs[stage]
        except IndexError:
            if line.startswith(stage_prefs[2:]):
                print(f'error in line {i+1}, stage {stage}: unexpected prefix: {line}')
                error = True
        else:
            if line.startswith(pref):
                if interface:
                    ii = len(res)
                    if len(interface) <= len(res):
                        print(f'error in {path}:{i+1}:stage{stage}: unexpected interface line: {line}')
                        error = True
                    else:
                        iline = interface[len(res)]
                        if line != iline:
                            print(f'error in {path}:{i+1}:stage{stage}: changed interface, expected: {iline}')
                            error = True
                res.append(line)
            else:
                print(f'error in {path}:{i+1}:stage{stage}: unexpected line: {line}')
                error = True
    if interface:
        for iline in interface[len(res):]:
            print(f'error in {path}:{i+1}:stage{stage}: missing line: {iline}')
            error = True
    if error:
        return None
    return interface or res
    
def build_shaders(files_meta):
    interfaces = {}
    for path, unused_attrs in files_meta.shader:
        if '.inc.' in path.name:
            continue
        interface = interfaces.get(path.suffix)
        interface = test_shader_interface(path, interface)
        interfaces[path.suffix] = interface
    assert all((v is not None) for v in interfaces.values())
    for path, attrs in files_meta.shader:
        yield from sync_shader(path, modify=True)
            
    
class FileLists:
    def __init__(self, filename):
        self.filename = filename
        self.allfiles = {}  # Path: Attributes
        self.includes = []  # Path
        self.excludes = []  # Path
        self.ignores = []  # str with filepattern
        self.vcs = None
        self.tempfiles = []  # Path
        self.lineno = 0
        self.errors = 0
        
    class Attributes:
        __slots__ = []
        __getitem__ = lambda self, key:         getattr(self, key)
        __setitem__ = lambda self, key, value:  setattr(self, key, value)
        __contains__ = lambda self, key:        hasattr(self, key)
        _eq_slot = lambda self, other, s: ((s in self) == (s in other) and (s in self and self[s]) == (s in other and other[s]))
        __eq__ = lambda self, other: all(self._eq_slot(other, s) for s in self.__slots__)
        def items(self):
            for k in self.__slots__:
                if k in self:
                    yield k, self[k]
        __str__ = __repr__ = lambda self:  '{}({})'.format(self.__class__.__name__, str({k:v for k,v in self.items() if v}))
        def update(self, items):
            for name, new_attrs in items.items():
                if name not in self:
                    self[name] = []
                old_attrs = self[name]
                for a in new_attrs:
                    if a not in old_attrs:
                        old_attrs.append(a)
                        
    def print_error(self, *args):
        print('{}:{}:'.format(self.filename, self.lineno+1), *args)
        self.errors += 1
        
    def ignored(self, filename):
        basename = filename.name
        if basename[0] == '.':
            return True
        for pattern in self.ignores:
            if Path(basename).match(pattern) or filename.match(pattern):
                return True
        return False
        
    def _inexclude(self, files, filename):
        if filename.is_file():
            if filename in self.allfiles:
                self.print_error('repeated:', filename)
            else:
                self.allfiles[filename] = self.Attributes()
            files.append(filename)
        else:
            self.print_error('not a file:', filename)
    def include(self, filename): self._inexclude(self.includes, filename)
    def exclude(self, filename): self._inexclude(self.excludes, filename)
    
    def glob_ne(self, pattern, *, notexists=False):
        filenames = sorted(Path(f) for f in iglob(pattern, recursive=True))
        if notexists:
            for f in filenames:
                self.print_error(f'file {str(f)!r} should only exist in vcs')
            return
        if not filenames:
            self.print_error('not found:', pattern)
        for f in filenames:
            if f.is_dir() or self.ignored(f):
                continue
            yield f
        
    def _parse_attr(self, attrs, a):
        name, values = a.split('=', 1) if '=' in a else ('flags',a)
        name = name.strip()
        if name != 'flags' and name in attrs:
            self.print_error('duplicate attribute name:', name)
        else:
            values = values.strip()
            if name not in attrs:
                attrs[name] = []
            if values:
                attrs[name] += [v.strip() for v in values.split(',')]
            
    def _parse_line(self, line, cmd, pattern, attrs):
        if not line.strip() or line[0] == '#':  # ignore empty line or comments
            return None,  cmd, pattern, attrs
        if line[0] != ' ':  # new command
            if cmd is not None:  # process the previous completed command
                return line,  cmd, pattern, attrs
            try:
                cmd, pattern = line.rstrip('\n').split(' ', 1)
            except ValueError:
                self.print_error('syntax error')
                return None,  None, None, None
            cmd, *attrs_list = cmd.split(':')
            attrs = {}
            for a in attrs_list:
                self._parse_attr(attrs, a)
            return None,  cmd, pattern, attrs or None
        else:  # attributes for command
            if cmd is None:
                self.print_error('unexpected indent')
                return None,  cmd, pattern, attrs
            if attrs is None:
                attrs = {}
            self._parse_attr(attrs, line.lstrip())
            return None,  cmd, pattern, attrs or None
            
    def _parse_cmd_ignores(self, cmd, pattern, flags):
        if cmd == 'ignore':
            if flags is None:
                self.ignores.append(pattern)
            else:
                self.print_error('flags not allowed for', repr(cmd))
            return True
        return False
        
    def _parse_cmd_vcs(self, cmd, pattern, flags):
        if cmd == 'vcs':
            if self.vcs is not None:
                self.print_error(f'command {cmd!r} allowed only once')
            elif flags is not None:
                self.print_error('flags not allowed for', repr(cmd))
            else:
                self.vcs = Path(pattern).exists()
            return True
        if self.vcs is None:
            self.vcs = True
        return False
        
    def on_file(self, filename, attrs):
        if not filename.is_file():
            self.print_error('not a file:', filename)
        if not 'flags' in attrs:
            return
        if attrs is not None:
            self.allfiles[filename].update(attrs)
        for flag in attrs['flags']:
            try:
                files = getattr(self, flag)
            except AttributeError:
                continue
        
    def _parse_cmd_inexclude(self, cmd, pattern, flags):
        assert flags != {}
        if cmd == 'include':
            for filename in self.glob_ne(pattern):
                self.include(filename)
                if flags is not None:
                    self.on_file(filename, flags)
        elif cmd == 'exclude':
            if flags is None:
                for filename in self.glob_ne(pattern, notexists=not self.vcs):
                    self.exclude(filename)
            else:
                self.print_error('flags not allowed for', repr(cmd))
        else:
            self.print_error('unknown command:', cmd)
            
    def _search_missing(self):
        for path, dirs, files in os.walk('.'):
            for filename in sorted(files):
                filename = Path(path, filename)
                if self.ignored(filename):
                    continue
                if filename not in self.allfiles:
                    self.print_error('missing:', filename)
            dirs[:] = [d for d in sorted(dirs) if not self.ignored(Path(path,d))]
            
    def _parse_cmd(self, stage, cmd, pattern, attrs):
        if stage == 1 and not self._parse_cmd_ignores(cmd, pattern, attrs):
            stage += 1
        if stage == 2 and not self._parse_cmd_vcs(cmd, pattern, attrs):
            stage += 1
        if stage == 3:
            self._parse_cmd_inexclude(cmd, pattern, attrs)
        return stage
        
    def parse(self):
        print('parsing', self.filename)
        
        with open(self.filename, 'rt', encoding='utf-8') as fin:
            stage = 1
            cmd = pattern = attrs = None
            for self.lineno, line in enumerate(fin):
                assert line is not None
                line, cmd, pattern, attrs = self._parse_line(line, cmd, pattern, attrs)
                while line is not None:
                    stage = self._parse_cmd(stage, cmd, pattern, attrs)
                    line, cmd, pattern, attrs = self._parse_line(line, None, None, None)
            if cmd is not None:
                self._parse_cmd(stage, cmd, pattern, attrs)
        self._search_missing()
        
class FileListsSetup (FileLists):
    def _mkfilesfilter(flag):
        def filesfilter(self):
            for fname, attrs in self.allfiles.items():
                if 'flags' in attrs and flag in attrs.flags:
                    yield fname, attrs
        return filesfilter
    data = property(_mkfilesfilter('data'))
    man = property(_mkfilesfilter('man'))
    po = property(_mkfilesfilter('po'))
    script = property(_mkfilesfilter('script'))
    py = property(_mkfilesfilter('py'))
    ext = property(_mkfilesfilter('ext'))
    depends = property(_mkfilesfilter('depends'))
    shader = property(_mkfilesfilter('shader'))
    
    class Attributes(FileLists.Attributes):
        __slots__ = 'flags depends variantgroup variantonly libs py2pyx language'.split()
        def __init__(self):
            self.variantgroup = []
            self.variantonly = []
            self.language = []
                
    def group_data(self):
        data_groups = {}
        for filename, _flags in self.data:
            _flags = _flags.flags
            install_dir = str(Path(*filename.parts[1:-1]))
            filename = str(filename)
            if install_dir in data_groups:
                files, flags = data_groups[install_dir]
                files.append(filename)
                flags += [f for f in _flags if f not in flags and f != 'data']
            else:
                data_groups[install_dir] = ([filename], [f for f in _flags if f != 'data'])
        return data_groups
        
    def get_modulesets(self):
        modulesets = {}
        _all_variants = all_variants()
        
        for filename, attrs in self.ext:
            modulename = get_modulename(filename)
            built_variants = []
            for variantflags in _all_variants:
                filtered_variantflags = filter_variant_by_groups(variantflags, attrs.variantgroup)
                modulevariantname = modulename + '_' + ''.join(filtered_variantflags)
                
                moduleset = modulesets.setdefault(variantflags, [])
                if any((v not in variantflags) for v in attrs.variantonly):
                    continue
                moduleset.append(modulevariantname)
                
                if filtered_variantflags in built_variants:
                    continue
                assert modulevariantname not in modulesets
                modulesets[modulevariantname] = moduleset
                built_variants.append(filtered_variantflags)
        return modulesets
    
    def gen_modulevariant_depends(self, filename, variantflags):
        modulename = get_modulename(filename)
        for filename_, attrs_ in self.allfiles.items():
            if 'depends' in attrs_ and modulename in attrs_.depends:
                if all(v in variantflags for v in attrs_.variantonly):
                    yield filename_
                    
    @staticmethod
    def pprint(obj):
        def asdict(o):
            if isinstance(o, (str, bool, int, type(None))):  return o
            if type(o) is dict:         return {asdict(k): asdict(v) for k,v in o.items()}
            if type(o) is list:         return [asdict(v) for v in o]
            if type(o) is tuple:        return tuple(asdict(v) for v in o)
            if hasattr(o, '__dict__'):  return asdict(o.__dict__)
            if hasattr(o, '__slots__') and o.__slots__:
                                        return {asdict(k): asdict(getattr(o, k)) for k in o.__slots__}
            if hasattr(o, '__str__'):  return str(o)
            assert False, type(o)
        import pprint
        pprint.pprint(asdict(obj))
        
    

