#!/bin/bash

#  Copyright © 2018-2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

set -e
#set -x

JOBS=-j6
export PYTHONVERSION=3.8.0b1
pythonversion=3.8
pythonbasedir=../trunk-deps/python
QTVERSION=5.12.4
export QT_SELECT=5
CONFIG="debug qml_debug"
all_archs="android_armv7 android_x86 desktop"
default_arch=desktop
pybiksrc="$(readlink -mn .)"
pythoninterpretersrcdir="interpreter-src_$PYTHONVERSION"
pythoninterpreterbuilddir="interpreter-build_$PYTHONVERSION"
pythoninterpreterinstalldir="interpreter-install_$PYTHONVERSION"
androidapi=22

function set_env() {
    ARCH="$1"
    qt_dir=~/.local/opt/Qt/$QTVERSION
    android_sdk=~/android-dev/Sdk
    if [ "$ARCH" = android_armv7 ]; then
        QMAKE=$qt_dir/android_armv7/bin/qmake
        QMAKE_spec="-spec android-clang"
        ANDROIDDEPLOYQT=$qt_dir/android_armv7/bin/androiddeployqt
        export ANDROID_NDK_PLATFORM=android-$androidapi
        export ANDROID_NDK_TOOLCHAIN_VERSION=4.9
        export ANDROID_NDK_HOST=linux-x86_64
        export ANDROID_NDK_ROOT=$android_sdk/ndk-bundle
        export ANDROID_SDK_ROOT=$android_sdk
        export ANDROID_HOME=$android_sdk
        export ANDROID_NDK_TOOLCHAIN_PREFIX=arm-linux-androideabi
        export ANDROID_NDK_TOOLS_PREFIX=arm-linux-androideabi
        ARCH_PLATFORM="$ANDROID_NDK_TOOLCHAIN_PREFIX"
    elif [ "$ARCH" = android_x86 ]; then
        QMAKE=$qt_dir/android_x86/bin/qmake
        QMAKE_spec="-spec android-clang"
        ANDROIDDEPLOYQT=$qt_dir/android_x86/bin/androiddeployqt
        export ANDROID_NDK_PLATFORM=android-$androidapi
        export ANDROID_NDK_TOOLCHAIN_VERSION=4.9
        export ANDROID_NDK_HOST=linux-x86_64
        export ANDROID_NDK_ROOT=$android_sdk/ndk-bundle
        export ANDROID_SDK_ROOT=$android_sdk
        export ANDROID_HOME=$android_sdk
        export ANDROID_NDK_TOOLCHAIN_PREFIX=x86
        export ANDROID_NDK_TOOLS_PREFIX=i686-linux-android
        ARCH_PLATFORM="$ANDROID_NDK_TOOLCHAIN_PREFIX"
    elif [ "$ARCH" = desktop ]; then
        QMAKE=$qt_dir/gcc_64/bin/qmake
        QMAKE_spec="-spec linux-g++"
        ARCH_PLATFORM=desktop
    else
        cmd_usage "unknown ARCH \"$ARCH\""
    fi
    
    pythonsrcdir="src-${ARCH_PLATFORM}_$PYTHONVERSION"
    pythonbuilddir="build-${ARCH_PLATFORM}_$PYTHONVERSION-Debug"
    pythoninstalldir="install-$ARCH_PLATFORM"
    
    pybikbuilddir=build-pybik-${ARCH}_Qt${QTVERSION//.}_Py${PYTHONVERSION//.}-Debug
    
    QMAKE_vars=("CONFIG+=$CONFIG" ARCH_PLATFORM=$ARCH_PLATFORM PYTHONVERSION=$PYTHONVERSION)
    ADB=$android_sdk/platform-tools/adb
}

function do_print_qmake_var() {
    mkdir temp_do_print_qmake_var
    cd temp_do_print_qmake_var
    cat >>print.pro <<'END'
message($$eval($$print))
END
    $QMAKE print.pro $QMAKE_spec print=$1 |& sed -n 's/^Project MESSAGE: //p'
    cd ..
    rm -r temp_do_print_qmake_var
}

function cmd_info() {
    echo "PYTHONVERSION: $PYTHONVERSION"
    echo "  pythonbasedir:               $pythonbasedir"
    echo "  pythonsrcdir:                $pythonsrcdir"
    echo "  pythonbuilddir:              $pythonbuilddir"
    echo "  pythoninstalldir:            $pythoninstalldir"
    echo "  pythoninterpretersrcdir:     $pythoninterpretersrcdir"
    echo "  pythoninterpreterbuilddir:   $pythoninterpreterbuilddir"
    echo "  pythoninterpreterinstalldir: $pythoninterpreterinstalldir"
    echo "pybikbuilddir: $pybikbuilddir"
    echo "  ANDROID_NDK_TOOLCHAIN_PREFIX: $ANDROID_NDK_TOOLCHAIN_PREFIX"
    echo "  DEFAULT_ANDROID_NDK_ROOT: $(do_print_qmake_var DEFAULT_ANDROID_NDK_ROOT)"
    echo "  ANDROID_NDK_ROOT:         $ANDROID_NDK_ROOT"
    echo "QMAKE: $QMAKE <project>.pro $QMAKE_spec ${QMAKE_vars[@]}"
    echo "  ARCH: $ARCH (android_armv7 android_x86 or desktop)"
    echo "  ARCH_PLATFORM: $ARCH_PLATFORM"
    echo "  QMAKESPEC:     $(do_print_qmake_var QMAKESPEC)"
    echo "  QMAKE_CC:      $(do_print_qmake_var QMAKE_CC)"
    echo "  QMAKE_CXX:     $(do_print_qmake_var QMAKE_CXX)"
}

function do_all() {
    local func=$1
    shift
    [ $# = 0 ] && { do_all $func $default_arch; return; }
    for ARCH in "$@"; do
        if [ "$ARCH" = all ]; then
            do_all $func $all_archs
        else
            set_env $ARCH
            $func
        fi
    done
}

function cmd_pybik_build() (
    test "$pybikbuilddir"
    [ -e "$pybikbuilddir/Makefile" ] || {
        cmd_pybik_qmake
    }
    cd "$pybikbuilddir"
    make $JOBS
)

function cmd_pybik_force() {
    cmd_pybik_full_clean
    cmd_pybik_build
}

function cmd_pybik_qmake() (
    test "$pybikbuilddir"
    mkdir -p "$pybikbuilddir"
    cd "$pybikbuilddir"
    $QMAKE "../freezer/pybik.pro" $QMAKE_spec "${QMAKE_vars[@]}"
    make qmake_all
)

function cmd_pybik_install() (
    test "$pybikbuilddir"
    cd "$pybikbuilddir"
    rm -rf android-build
    make INSTALL_ROOT=android-build install
    [ "$ANDROIDDEPLOYQT" ] && {
        $ANDROIDDEPLOYQT --input android-libpybik.so-deployment-settings.json \
                         --output android-build \
                         --deployment bundled --android-platform $ANDROID_NDK_PLATFORM \
                         --jdk /usr/lib/jvm/default-java --verbose --gradle
    }
)

function cmd_pybik_deploy() (
    test "$ANDROIDDEPLOYQT"
    test "$pybikbuilddir"
    cd "$pybikbuilddir"
    if [ "$ARCH" = android_armv7 ]; then
        DEVICE="$($ADB -d get-serialno)"
    elif [ "$ARCH" = android_x86 ]; then
        DEVICE="$($ADB -e get-serialno)"
    fi
    test "$DEVICE"
    echo "device: $DEVICE"
    $ANDROIDDEPLOYQT --verbose --output android-build --no-build \
                     --input android-libpybik.so-deployment-settings.json \
                     --deployment bundled --gradle --reinstall --device $DEVICE
    $ADB -s $DEVICE pull /system/bin/app_process32 app_process
    $ADB -s $DEVICE pull /system/lib/libc.so libc.so
    $ADB -s $DEVICE shell am start -n pybik.de.barcc.pybik.app/org.qtproject.qt5.android.bindings.QtActivity
)

function cmd_pybik_clean() (
    test "$pybikbuilddir"
    cd "$pybikbuilddir"
    make clean
)

function cmd_pybik_full_clean() (
    test "$pybikbuilddir"
    rm -rf "$pybikbuilddir"/android-build
    shopt -s dotglob
    rm -f "$pybikbuilddir"/*
)

function cmd_python_build() (
    test "$pythonbuilddir"
    [ -e "$pythonbuilddir/Makefile" ] || {
        cmd_python_qmake
    }
    cd "$pythonbuilddir"
    make $JOBS
    make $JOBS install
)

function cmd_python_force() {
    cmd_python_full_clean
    cmd_python_build
}

function cmd_python_qmake() (
    test "$pythonsrcdir"
    test "$pythonbuilddir"
    mkdir -p "$pythonbuilddir"
    cd "$pythonbuilddir"
    $QMAKE -o Makefile "../$pythonsrcdir/python.pro" $QMAKE_spec "${QMAKE_vars[@]}"
    make qmake_all
)

function cmd_python_clean() (
    test "$pythonbuilddir"
    cd "$pythonbuilddir"
    make clean
)

function cmd_python_full_clean() {
    test "$pythonbuilddir"
    rm -f "$pythonbuilddir"/*
}

function cmd_python_interpreter() (
    test "$pythoninterpreterbuilddir"
    test "$pythoninterpreterinstalldir"
    mkdir -p "$pythoninterpreterbuilddir"
    mkdir -p "$pythoninterpreterinstalldir"
    prefix="$(readlink -en "$pythoninterpreterinstalldir")"
    cd "$pythoninterpreterbuilddir"
    "../$pythoninterpretersrcdir"/configure --prefix="$prefix"
    make $JOBS
    make $JOBS install
)

function do_python_get_src() {
    [ -d "$1" ] && rm -r "$1"
    [ -d "Python-$PYTHONVERSION" ] && rm -r "Python-$PYTHONVERSION"
    echo "extract xz-archive/Python-$PYTHONVERSION.tar.xz"
    tar --extract -f xz-archive/Python-$PYTHONVERSION.tar.xz
    test -d "Python-$PYTHONVERSION"
    mv "Python-$PYTHONVERSION" "$1"
    echo "Python source in $1"
}

function do_python_configure() (
    test "$pythoninterpreterinstalldir"
    test "$pythonsrcdir"
    python_dir="$(readlink -e "$pythoninterpreterinstalldir/bin")"
    cd "$pythonsrcdir"
    export CC=$(do_print_qmake_var QMAKE_CC)
    export CXX=$(do_print_qmake_var QMAKE_CXX)
    export AR=$(do_print_qmake_var QMAKE_AR)
    export RANLIB=$(do_print_qmake_var QMAKE_RANLIB)
    export READELF=$(do_print_qmake_var CROSS_COMPILE)-readelf
    export ac_cv_file__dev_ptmx=no
    export ac_cv_file__dev_ptc=no
    export ac_cv_have_long_long_format=yes
    PATH=$python_dir:$PATH
    export PYTHON_FOR_REGEN=$python_dir/python3
    
    ./configure --host=$ANDROID_NDK_TOOLS_PREFIX --build=x86_64-linux-gnu --prefix=/usr \
                --disable-shared --disable-ipv6 --without-ensurepip
    echo "copy pyconfig.h"
    test -f pyconfig.h  #XXX: file must exist before qmake generates list of installed files
    sleep 1             #XXX: ^ needs really older timestamp
    cp pyconfig.h pyconfig.h.orig
)

function cmd_python_get_src() (
    test "$pythonsrcdir"
    test "$pybiksrc"
    do_python_get_src "$pythonsrcdir"
    do_python_configure
    cd "$pythonsrcdir"
    ln -s "$pybiksrc/freezer" freezer
    ln -s "$pybiksrc/buildlib/gentools.py" gentools.py
    ln -s "$pybiksrc/freezer/python.pro" python.pro
    cp Modules/config.c.in Modules/config.c
    if [ -f "$pybiksrc/freezer/python-$pythonversion.patch" ]; then
        patch -p1 <"$pybiksrc/freezer/python-$pythonversion.patch"
    fi
)

function cmd_python_get_interpreter() {
    test "$pythoninterpretersrcdir"
    do_python_get_src "$pythoninterpretersrcdir"
}

function cmd_python_download_source() (
    mkdir -p "xz-archive"
    cd "xz-archive"
    if [ "$1" ]; then
        _PYTHONVERSION="$1"
        shift
    else
        _PYTHONVERSION="$PYTHONVERSION"
    fi
    [ $# = 0 ] || cmd_usage 'to many args'
    pyversion=$(sed 's/[a-z].*//' <<<"$_PYTHONVERSION")
    test $pyversion
    if [ -e "Python-$_PYTHONVERSION.tar.xz" ]; then
        echo "Python-$_PYTHONVERSION.tar.xz already exists"
    else
        wget "https://www.python.org/ftp/python/$pyversion/Python-$_PYTHONVERSION.tar.xz"
    fi
)

function cmd_bootstrap() {
    (   test "$pythonbasedir"
        mkdir -p "$pythonbasedir"
        cd "$pythonbasedir"
        cmd_python_download_source
        cmd_python_get_interpreter
        cmd_python_interpreter
        do_all cmd_python_get_src "$@"
        do_all cmd_python_build "$@"
    )
    do_all cmd_pybik_build "$@"
}


function cmd_usage() {
    [ "$1" ] && echo "error: $1"
    prog="$(basename $0)"
    echo "$prog info ARCH"
    echo "-----------------------------------------"
    echo "$prog pybik ARCH"
    echo "$prog pybik force ARCH"
    echo "$prog pybik qmake ARCH"
    echo "$prog pybik install ARCH"
    echo "$prog pybik deploy ARCH"
    echo "$prog pybik clean ARCH"
    echo "$prog pybik clean-full ARCH"
    echo "-----------------------------------------"
    echo "$prog python ARCH"
    echo "$prog python force ARCH"
    echo "$prog python qmake ARCH"
    echo "$prog python clean ARCH"
    echo "$prog python clean-full ARCH"
    echo "$prog python interpreter"
    echo "-----------------------------------------"
    echo "$prog python source ARCH"
    echo "$prog python source-interpreter"
    echo "$prog python download-source [VERSION]"
    echo "-----------------------------------------"
    echo "$prog bootstrap"
    echo "-----------------------------------------"
    echo "ARCH: $all_archs all"
    exit 1
}

[ $# = 0 ] && cmd_usage 'missing arg 1'
case "$1" in
    info)
        shift; do_all cmd_info "$@";;
    pybik)
        shift
        case "$1" in
            force)      shift; do_all cmd_pybik_force "$@";;
            qmake)      shift; do_all cmd_pybik_qmake "$@";;
            install)    shift; do_all cmd_pybik_install "$@";;
            deploy)     shift; do_all cmd_pybik_deploy "$@";;
            clean)      shift; do_all cmd_pybik_clean "$@";;
            clean-full) shift; do_all cmd_pybik_full_clean "$@";;
            *)                 do_all cmd_pybik_build "$@";;
        esac;;
    python)
        shift
        test "$pythonbasedir"
        cd "$pythonbasedir"
        case "$1" in
            force)      shift; do_all cmd_python_force "$@";;
            qmake)      shift; do_all cmd_python_qmake "$@";;
            clean)      shift; do_all cmd_python_clean "$@";;
            clean-full) shift; do_all cmd_python_full_clean "$@";;
            interpreter)shift; [ $# = 0 ] || cmd_usage 'to many args'; cmd_python_interpreter;;
            source)     shift; do_all cmd_python_get_src "$@";;
            source-interpreter)
                        shift; [ $# = 0 ] || cmd_usage 'to many args'; cmd_python_get_interpreter;;
            download-source)
                        shift; cmd_python_download_source "$@";;
            *)                 do_all cmd_python_build "$@";;
        esac;;
    bootstrap)  shift; cmd_bootstrap "$@";;
    *)  cmd_usage "unknown \"$1\"";;
esac

