//  Copyright © 2017-2019  B. Clausius <barcc@gmx.de>
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <stdio.h>

#include <Python.h>

#include <QtCore/QtDebug>

#include <QByteArray>
#include <QString>
#include <QTextCodec>

#ifdef Q_OS_ANDROID
#include <android/log.h>
#endif


static QTextCodec *locale_codec;

extern struct _inittab _PyImport_InittabModules[];
extern struct _frozen _PyImport_FrozenModules[];


#ifdef Q_OS_ANDROID
void android_message_handler(QtMsgType type, const QMessageLogContext &context, const QString &message) {
    android_LogPriority priority = ANDROID_LOG_INFO;
    switch (type) {
        case QtDebugMsg:    priority = ANDROID_LOG_DEBUG; break;
        case QtInfoMsg:     priority = ANDROID_LOG_INFO; break;
        case QtWarningMsg:  priority = ANDROID_LOG_WARN; break;
        case QtCriticalMsg: priority = ANDROID_LOG_ERROR; break;
        case QtFatalMsg:    priority = ANDROID_LOG_FATAL; break;
    };
    __android_log_print(priority, "pybik/Qt", "%s", qPrintable(message));
}
#endif

int get_env_inspect() {
    //FIXME: only debug builds
    char *p = Py_GETENV("PYTHONINSPECT");
    return (p != 0) && (*p != '\0');
}

int init_locale()
{
    locale_codec = QTextCodec::codecForLocale();
    if (!locale_codec) {
        qCritical("error: locale codec not found");
        return 1;
    }
    
    if (Py_FileSystemDefaultEncoding)
        return 0;
        
    static QByteArray name;
    name = locale_codec->name();
    if (name != "System") {
        Py_FileSystemDefaultEncoding = name.data();
        Py_HasFileSystemDefaultEncoding = 1;
    }
    return 0;
}

int init_modules()
{
    // frozen modules
    PyImport_FrozenModules = _PyImport_FrozenModules;
    // extension modules
    if (PyImport_ExtendInittab(_PyImport_InittabModules) < 0) {
        qCritical("error: insufficient memory for built-in modules");
        return 1;
    }
    return 0;
}

wchar_t **convert_argv(int argc, char **argv)
{
    wchar_t **w_argv = new wchar_t *[argc + 1];
    for (int i = 0; i < argc; i++) {
        QString qs_arg = locale_codec->toUnicode(argv[i]);
        wchar_t *w_arg = new wchar_t[qs_arg.length() + 1];
        w_arg[qs_arg.toWCharArray(w_arg)] = 0;
        w_argv[i] = w_arg;
    }
    w_argv[argc] = NULL;
    return w_argv;
}

int run_main_script()
{
    int res;
    res = PyImport_ImportFrozenModule("__main__");
    if (res == 0) {
        qCritical("error: __main__ not frozen");
        return 1;
    }
    if (res < 0) {
        PyErr_Print();
        return 1;
    }
    return 0;
}

int main(int argc, char **argv)
{
#ifdef Q_OS_ANDROID
    qInstallMessageHandler(android_message_handler);
#endif
    
    int inspect = get_env_inspect();
    
    // Python globals
    Py_FrozenFlag = 1;
    Py_NoSiteFlag = 1;
    Py_IgnoreEnvironmentFlag = 1;
    Py_OptimizeFlag = 1;
    
    if (init_locale())
        return 1;
        
    if (init_modules())
        return 1;
        
    wchar_t **w_argv = convert_argv(argc, argv);
        
    if (argc >= 1)
        Py_SetProgramName(w_argv[0]);
    Py_Initialize();
    PySys_SetArgvEx(argc, w_argv, 0);
    
    int res = run_main_script();
    
    if (inspect && isatty((int)fileno(stdin)))
        res = PyRun_AnyFile(stdin, "<stdin>") != 0;
        
    Py_Finalize();
    return res;
}

