#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013, 2016-2017, 2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
import re
from contextlib import suppress
import subprocess
from glob import glob
from pathlib import Path

try:
    from send_tracebacks import send_dbus_message
except ImportError:
    send_dbus_message = lambda *args: print(*args)

import buildlib.utils

open_rt = lambda filename: open(filename, 'rt', encoding='utf-8')
open_wt = lambda filename: open(filename, 'wt', encoding='utf-8')


def convert_plugin_for_translation(source_filename, target_filename):
    from pybiklib import pluginlib
    plugin = pluginlib.PluginFileLoader(str(source_filename))
    source_filename = str(source_filename.absolute())
    with target_filename.open('wt', encoding='utf-8') as target:
        translation_comment = None
        dstlineno = 0
        srclineno = 0
        for paragraph in plugin.body:
            path, pathitems, srclineno, translation_comment = paragraph['Path']
            if translation_comment is not None:
                if ':' in translation_comment:
                    translation_key, translation_comment = translation_comment.strip('#').strip().split(':', 1)
                    if translation_key:
                        translation_comment = '#TRANSLATORS: ' + translation_comment.strip()
                    else:
                        translation_comment = None
                        send_dbus_message('Warning', 'Empty translation key', source_filename, srclineno)
                else:
                    translation_key = translation_comment.strip('#').strip().split()
                    if translation_key:
                        translation_key = translation_key[0]
                        translation_comment = '#TRANSLATORS: ' + translation_comment.strip('#').strip()
                    else:
                        translation_comment = None
                        send_dbus_message('Warning', 'Empty translation key', source_filename, srclineno)
                if ' ' in translation_key:
                    send_dbus_message('Warning', 'Translation key contains whitespace', source_filename, srclineno)
            if pathitems is not None:
                values = []
                for unused, v in pathitems:
                    if v:
                        if translation_comment and translation_key in re.split(r'\W+', v):
                            values.append(translation_comment)
                        values.append(v)
                while dstlineno < srclineno:
                    target.write('\n')
                    dstlineno += 1
                for v in values:
                    target.write(v)
                    target.write('\n')
                    dstlineno += 1
                translation_comment = None
            else:
                translation_comment = None
                continue
                
class FileListsPotfiles (buildlib.utils.FileLists):
    def on_file(self, filename, attrs):
        for aname, flags in attrs.items():
            if aname != 'flags':
                self.print_error('unknown attribute:', aname)
            elif len(flags) != 1:
                self.print_error('only one flag allowed')
            elif flags[0] == 'pybikplugin':
                self.on_file_pybikplugin(filename)
            elif flags[0] == 'desktop':
                self.on_file_desktop(filename)
            elif flags[0] == 'qtui':
                self.on_file_qtui(filename)
            else:
                self.print_error('unknown flag:', flags[0])
            
    def on_file_pybikplugin(self, filename):
        tempfile = filename.with_name(filename.name + '.py')
        print(filename, '->', tempfile)
        convert_plugin_for_translation(filename, tempfile)
        self.tempfiles.append(tempfile)
        self.includes.remove(filename)
        self.excludes.append(filename)
        self.include(tempfile)
        
    def on_file_desktop(self, filename):
        from buildlib.create_docs import create_doc
        if filename.name.endswith('.desktop.in'):
            desktop_dir = Path('data/app-meta')
            if not desktop_dir.exists():
                print('creating directory:', desktop_dir)
                desktop_dir.mkdir()
            tempfile = desktop_dir / filename.name
            template = filename.with_suffix('')
            print(filename, '->', tempfile)
            
            create_doc(template, tempfile)
            #XXX: prevent translation of Icon
            text = tempfile.read_bytes()
            tempfile.write_bytes(text.replace(b'Icon=pybik', b''))
            
            self.tempfiles.append(tempfile)
            self.includes.remove(filename)
            self.excludes.append(filename)
            self.include(tempfile)
        else:
            self.print_error('no desktop file template:', filename)
            
    def on_file_qtui(self, filename):
        tempfile = filename.with_name(filename.name + '.h')
        print(filename, '->', tempfile)
        
        subprocess.call(['uic', filename, '-o', tempfile, '--translate', 'gettext_translate'])
        text = tempfile.read_bytes()
        text = re.sub(rb'gettext_translate(.*), Q_NULLPTR', rb'gettext\1', text)
        tempfile.write_bytes(text)
        
        self.tempfiles.append(tempfile)
        self.includes.remove(filename)
        self.excludes.append(filename)
        self.include(tempfile)
        
def generate_potfiles_in():
    os.environ['QT_SELECT'] = '5'
    result = FileListsPotfiles('po/POTFILES.in.in')
    result.parse()
    
    with open_wt('po/POTFILES.in') as fout:
        for filename in result.includes:
            fout.write(str(filename) + '\n')
    result.tempfiles.append(Path('po/POTFILES.in'))
            
    print()
    return result.tempfiles, result.errors
    
def cmd_merge_po():
    '''Update po(t) files and print a report'''
    from pybiklib import config
    from pybiklib import model
    from pybiklib import pluginlib
    model.Model.load_index()
    pluginlib._ = lambda v: '_({!r})'.format(v)
    pluginlib.ngettext = lambda s,p,n: 'ngettext({!r}, {!r}, {})'.format(s,p,n).replace('{','{{').replace('}','}}')
    
    wd = os.getcwd()
    temp_files, errors = generate_potfiles_in()
    try:
        if errors:
            print(errors, 'errors.')
            return
        subprocess.call('xgettext --width=79 --package-name={0.PACKAGE} --package-version={0.VERSION} '
            '--msgid-bugs-address={0.CONTACT_FILEBUG} --keyword=N_ --from-code=utf-8 '
            '--add-comments=TRANSLATORS: '
            '-o po/{0.PACKAGE}.pot -f po/POTFILES.in'
            .format(config).split())
        with open_rt('po/{0.PACKAGE}.pot'.format(config)) as file:
            text = file.read()
        with open_wt('po/{0.PACKAGE}.pot'.format(config)) as file:
            file.write(text.replace("Copy" "right (C) YEAR THE PACKAGE'S COPY" "RIGHT HOLDER",
                                    '{0.COPYRIGHT} <{0.CONTACT_EMAIL}>'.format(config))
                           .replace('#. TRANSLATORS:', '#.')
                      )
    finally:
        for temp_file in temp_files:
            temp_file.unlink()
            
def blacklisted(langcode, href, po_href):
    #XXX: Blacklist hack
    if href != 'https://launchpad.net/~barcc':
        return False
    if langcode not in ['de', 'en_GB']:
        return True
    if href != po_href != None:
        return True
    return False
    
class POEntry:
    def __init__(self):
        self.po_empty_a = []
        self.po_comments_t = []  #  translator comments (starts with whitespace)
        self.po_comment_l = {}  #  comment last translator/reviewer (starts with ' translated by: ' and ' reviewed by: ')
        self.po_comments_s = []  #  special comments (starts with .:,|)
        self.po_msg = []  # list of ('msg*', lines)
        self.po_empty_e = None
        
    def __repr__(self):
        return 'POEntry({!r})'.format(self.__dict__)
    def lines(self):
        yield from self.po_empty_a
        yield from self.po_comments_t
        for ptype in 'translated by', 'reviewed by':
            po_comment = self.po_comment_l.get(ptype)
            if po_comment is not None:
                yield po_comment
        yield from self.po_comments_s
        for msgtype, msglines in self.po_msg:
            msglines = iter(msglines)
            yield ' '.join((msgtype, next(msglines)))
            yield from msglines
        yield self.po_empty_e
        
    def persons(self):
        for ptype, line in self.po_comment_l.items():
            assert line.startswith('# %s: '%ptype), (ptype, line)
            po_person, po_date = line[len(ptype)+4:].rsplit(' on ', 1)
            po_person, po_href = po_person.rsplit(' ', 1)
            yield ptype, po_person, po_href, po_date
    def msgid(self):
        if not self.po_msg:
            return None
        assert self.po_msg[0][0] == 'msgid'
        return eval(' '.join(self.po_msg[0][1]))
    def has_msgid(self):
        return self.po_msg and self.po_msg[0][0] == 'msgid'
    def is_translated(self):
        return any(len(msglines) > 1 or msglines[0] != '""'
                   for msgtype, msglines in self.po_msg
                   if msgtype.startswith('msgstr'))
    def is_obsolete(self):
        return self.po_comments_s and self.po_comments_s[-1].startswith('#~ ')
        
def read_pofile(langcode):
    with open('po/%s.po'%langcode, 'rt', encoding='utf-8') as f:
        lines = f.read().split('\n')
    return parse_pofile(lines)
    
def parse_pofile(lines):
    po_messages = []
    po_entry = POEntry()
    for line in lines:
        if not line:
            if not po_entry.po_msg and not po_entry.po_comments_s:
                assert not po_entry.po_comments_t
                # empty lines at start and duplicate empty lines
                po_entry.po_empty_a.append(line)
                continue
            assert po_entry.has_msgid() or po_entry.is_obsolete(), repr(po_entry)
            # next entry
            po_entry.po_empty_e = line
            po_messages.append(po_entry)
            po_entry = POEntry()
        elif line.startswith('# translated by: '):
            assert not po_entry.po_msg
            po_entry.po_comment_l['translated by'] = line
        elif line.startswith('# reviewed by: '):
            assert not po_entry.po_msg
            po_entry.po_comment_l['reviewed by'] = line
        elif line.startswith('# '):
            assert not po_entry.po_msg
            po_entry.po_comments_t.append(line)
        elif line.startswith('#'):
            assert not po_entry.po_msg
            po_entry.po_comments_s.append(line)
        elif line.startswith('msg'):
            msg_type, msg_line = line.split(None, 1)
            po_entry.po_msg.append((msg_type, [msg_line]))
        else:  # continuation lines for msg* lines
            assert len(line) >= 2 and line[0] == '"' == line[-1], repr(line)
            po_entry.po_msg[-1][1].append(line)
    assert not po_entry.po_comments_t and not po_entry.po_comment_l
    assert not po_entry.po_comments_s and not po_entry.po_msg
    return po_messages
    
def match_translations(messages, po_messages):
    po_message_number = {e.msgid(): i-1 for i, e in enumerate(po_messages)}
    if None in po_message_number:
        del po_message_number[None]
    message_reorder = {}
    for i, message, persons in messages:
        assert type(i) is int, i
        assert type(message) is str, (type(message), message)
        po_i = po_message_number.get(message)
        if po_i != i:
            message_reorder[i] = po_i
    return message_reorder
    
def merge_message(langcode, po_entry, persons):
    if not po_entry.is_translated():
        po_entry.po_comment_l.clear()
        persons.clear()
        return
    for ptype in 'translated by', 'reviewed by':
        person_info = persons.get(ptype)
        if person_info is None:
            continue
        person, href, date = person_info
        po_comment = po_entry.po_comment_l.get(ptype)
        if po_comment is None:
            if blacklisted(langcode, href, None):
                del persons[ptype]
            else:
                po_entry.po_comment_l[ptype] = '# {}: {} {} on {}'.format(ptype, person, href, date)
        else:
            assert po_comment.startswith('# %s: '%ptype), po_comment
            po_person, po_date = po_comment[len(ptype)+4:].rsplit(' on ', 1)
            po_person, po_href = po_person.rsplit(' ', 1)
            if blacklisted(langcode, href, po_href):
                persons[ptype] = po_person, po_href, po_date
            elif href == po_href:
                po_entry.po_comment_l[ptype] = '# {}: {} {} on {}'.format(ptype, person, href, po_date)
                persons[ptype] = person, href, po_date
            else:
                po_entry.po_comment_l[ptype] = '# {}: {} {} on {}'.format(ptype, person, href, date)
                
def merge_translators_pofile(langcode, count, messages, po_messages):
    assert len(messages) <= count
    result = []
    po_persons = {}
    po_count = 0
    po_translated = 0
    m_i = 0
    for po_i, po_entry in enumerate(po_messages):
        assert po_entry.has_msgid() or po_entry.is_obsolete()
        if po_i > 0:  # keep header unmodified
            if m_i < len(messages):
                i, message, persons = messages[m_i]
                assert po_i <= i+1 <= count, (po_i, i+1, count)
            else:
                i = po_i
                assert po_i < i+1, (po_i, i+1, count)
            if po_i == i+1:
                # found message that can be merged
                merge_message(langcode, po_entry, persons)
                m_i += 1
            if po_entry.has_msgid():
                for po_person, po_href in set((p,h) for t,p,h,d in po_entry.persons()):
                    try:
                        po_person_, po_href_, po_person_count_  = po_persons[po_href]
                    except KeyError:
                        po_persons[po_href] = po_person, po_href, 1
                    else:
                        assert (po_person_, po_href_) == (po_person, po_href), ((po_person_, po_href_), (po_person, po_href))
                        po_persons[po_href] = po_person, po_href, po_person_count_+1
                if po_entry.is_translated():
                    po_translated += 1
                po_count += 1
        result.extend(po_entry.lines())
        
    assert po_count == count, (po_count, count)
    po_persons = sorted(po_persons.values(), key=lambda v: (-v[2], v[0], v[1]))
    for po_person, po_href, po_person_count in po_persons:
        assert po_person_count <= po_count, (po_person, po_href, po_person_count, po_count)
    po_persons = [(p,h) for p,h,c in po_persons]
    with open('po/%s.po'%langcode, 'wt', encoding='utf-8') as f:
        f.write('\n'.join(result))
    return po_translated == po_count, po_persons
            
def pprint_translators(translation_authors, file):
    linguas = buildlib.utils.get_linguas()
    
    print('# generated with:', *sys.argv, file=file)
    print(file=file)
    print('translators = [', file=file)
    for langname, langcode, complete, persons in translation_authors:
        if langcode not in linguas:
            continue
        print('  ({!r}, {!r}, {!r}, ['.format(langname, langcode, complete), file=file)
        for person in persons:
            print('    {!r},'.format(person), file=file)
        print('  ]),', file=file)
    print(']\n', file=file)
    
def read_files_meta():
    global files_meta
    try:
        files_meta
    except NameError:
        filename = 'files-meta'
        files_meta = buildlib.utils.FileListsSetup(filename)
        files_meta.parse()
        if files_meta.errors:
            print(f'{files_meta.errors} errors in "{filename}"')
            sys.exit(1)
            
def cmd_merge_translators():
    import bc_lp_project_translators  #XXX: not at module level, keep README.bzr working without bc-dev-tool
    
    cache_dir='temp/translation_pages'
    kwargs = dict(
                cache_dir=cache_dir,
                #force_download=True,
            )
    read_files_meta()
            
    print('=== Collect translated message numbers ===')
    poinfo = {}
    for filename, unused_attrs in files_meta.po:
        langcode = filename.stem
        print(langcode, end=' ', flush=True)
        lines = filename.read_text(encoding='utf-8').split('\n')
        entries = parse_pofile(lines)
        numbers = [i-1 for i, e in enumerate(entries) if i>0 and e.is_translated()]
        poinfo[langcode] = len(entries)-1, numbers
    kwargs['poinfo'] = poinfo
    print()
    
    print('=== Collect info about translators ===')
    translation_authors = bc_lp_project_translators.get_translation_authors_full('pybik', **kwargs)
    
    print('=== Update last translators in pofiles ===')
    po_authors = []
    for unused_template, langname, langcode, count, messages in translation_authors:
        print(end=langcode, flush=True)
        try:
            po_messages = read_pofile(langcode)
        except FileNotFoundError:
            continue
        message_reorder = match_translations(messages, po_messages)
        if message_reorder:
            print(end='?')
            bc_lp_project_translators.reorder_message_in_cache(cache_dir, langcode, messages, message_reorder)
        print(end=' ', flush=True)
        po_complete, po_persons = merge_translators_pofile(langcode, count, messages, po_messages)
        po_authors.append((langname, langcode, po_complete, po_persons))
    print()
        
    print('=== Generate po/LINGUAS ===')
    buildlib.utils.update_linguas(files_meta.po)
    
    print('=== Generate pybiklib/translators.py ===')
    with open('pybiklib/translators.py', 'wt', encoding='utf-8') as file:
        pprint_translators(po_authors, file)
        
    print('=== Generate pybiklib/languages.py ===')
    with open('pybiklib/languages.py', 'wt', encoding='utf-8') as file:
        buildlib.utils.generate_all_lang_translations(file)
    
def cmd_revert_translators():
    read_files_meta()
    print('=== Revert translator names ===')
    for filename, unused_attrs in files_meta.po:
        langcode = filename.stem
        print(langcode, end=' ', flush=True)
        p = subprocess.run(['bzr', 'cat', filename],
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        if p.stderr:
            print()
            print(p.stderr)
        if p.returncode:
            continue
        lines = p.stdout.split('\n')
        po_messages = parse_pofile(lines)
        translators = {e.msgid(): e.po_comment_l for e in po_messages if e.msgid() is not None}
        lines = filename.read_text(encoding='utf-8').split('\n')
        po_messages = parse_pofile(lines)
        result = []
        for po_entry in po_messages:
            po_comment_l = translators.get(po_entry.msgid())
            if po_comment_l is not None:
                po_entry.po_comment_l = po_comment_l
            result.extend(po_entry.lines())
        filename.write_text('\n'.join(result), encoding='utf-8')
    print()
    
def cmd_clear_outdated_cache():
    import dateutil.parser
    import bc_lp_project_translators  #XXX: not at module level, keep README.bzr working without bc-dev-tool
    
    read_files_meta()
    poinfo = {}
    for filename, unused_attrs in files_meta.po:
        langcode = filename.stem
        print(' ', end=langcode, flush=True)
        p = subprocess.run(['bzr', 'log', '-l', '1', filename], stdout=subprocess.PIPE, universal_newlines=True)
        if p.returncode:
            continue
        date = [l for l in p.stdout.split('\n') if l.startswith('timestamp: ')][0].split(None, 1)[1]
        date = dateutil.parser.parse(date)
        entries = read_pofile(langcode)
        entries = [i for i, e in enumerate(entries) if i>0 and e.is_translated()]
        poinfo[langcode] = (date, entries)
    print()
    cache_dir='temp/translation_pages'
    count = bc_lp_project_translators.clear_outdate_cache(cache_dir, poinfo)
    print(count, 'Dateien gelöscht')
    
def cmd_update_docs():
    from buildlib.create_docs import create_doc
    print('creating README')
    create_doc('README')
    print('creating INSTALL')
    create_doc('INSTALL')
    print('creating net.launchpad.pybik.appdata.xml.in')
    create_doc('data/templates/net.launchpad.pybik.appdata.xml.in')
    print('creating copyright')
    create_doc('copyright', skip='deb')
    print('creating debian/copyright')
    create_doc('copyright', 'debian/copyright', skip='upstream')
    print('creating debian/control')
    create_doc('debian/control')
    print('creating debian/upstream/metadata')
    create_doc('debian/upstream/metadata')
    
def cmd_update_linguas():
    read_files_meta()
    buildlib.utils.update_linguas(files_meta.po)

def cmd_full_update():
    cmd_update_docs()
    cmd_merge_po()
    cmd_merge_translators()
    
def cmd_merge_translations_branch():
    p = subprocess.run(['bzr', 'merge'])
    if p.returncode:
        return
    cmd_revert_translators()
    cmd_merge_translators()
    
def cmd_environment():
    path = Path.home() / '.local/opt/Qt'
    for p in sorted(path.iterdir()):
        p = p / "gcc_64/lib"
        if p.is_dir():
            print(f'export LD_LIBRARY_PATH={p}')
    path = Path('../trunk-deps/python')
    for p in sorted(path.glob('interpreter-install_*')):
        if p.is_dir():
            print(f'alias py3={p/"bin/python3"}')
            
def _run_emulator(name):
    emulator = Path('~/android-dev/Sdk/emulator/emulator').expanduser().resolve()
    subprocess.run(['sudo', emulator, name])
    
def cmd_emulator_x86_android_22():
    _run_emulator('@x86-android-22')
    
def cmd_emulator_x86_android_23():
    _run_emulator('@x86-android-23')
    
def cmd_emulator_x86_android_26():
    _run_emulator('@x86-android-26')
    

accepted_commands = {k.split('_',1)[1].replace('_','-'):v for k,v in globals().items() if k.startswith('cmd_')}

def usage():
    print('usage:')
    prog = os.path.basename(__file__)
    print(' ', prog, '-h|--help')
    for cmd in accepted_commands:
        print(' ', prog, cmd)
    sys.exit(1)
    
def main():
    args = sys.argv[1:]
    if len(args) != 1:
        usage()
    elif args[0] in ['-h', '--help']:
        usage()
    elif args[0] in accepted_commands:
        accepted_commands[args[0]]()
    else:
        usage()
    
if __name__ == '__main__':
    main()
    

