//  Copyright © 2019  B. Clausius <barcc@gmx.de>
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include <QtCore/QtGlobal>
#include <QtCore/QtDebug>
#include <QtCore/QUrl>

#include <QtGui/QPainter>

#include <QtQuick/QQuickWindow>

#include <Qt3DCore/QEntity>
#include <Qt3DCore/QTransform>
#include <Qt3DRender/QAttribute>
#include <Qt3DRender/QBuffer>
#include <Qt3DRender/QGeometryRenderer>
#include <Qt3DRender/QGeometry>
#include <Qt3DRender/QMaterial>
#include <Qt3DRender/qfilterkey.h>
#include <Qt3DRender/qeffect.h>
#include <Qt3DRender/qtechnique.h>
#include <Qt3DRender/qshaderprogram.h>
#include <Qt3DRender/qparameter.h>
#include <Qt3DRender/qrenderpass.h>
#include <Qt3DRender/qgraphicsapifilter.h>
#include <Qt3DRender/QPaintedTextureImage>
#include <Qt3DRender/QTexture>


using namespace Qt3DCore;
using namespace Qt3DRender;

//pxd>from qt cimport *
//pxd>from qtq cimport *

//pxd>cdef enum: #
enum {
//pxd:
    DEBUG_SHADERGL3 = 0x10000,
    DEBUG_SHADERGL2 = 0x20000,
    DEBUG_SHADERES2 = 0x40000,
    DEBUG_SHADER_MASK = 0x70000,
//px.
};
extern long debug;

//pxd>cdef extern from "qt_impl.h":#
//pxm>DINDENT

class LabelTextureImage : public QPaintedTextureImage
{
public:
    void paint(QPainter *painter) override;
};

class PerVertexColorMaterial
{
public:
    QMaterial *material;
    QEffect *vertexEffect;
    LabelTextureImage labeltextureimage;
    QTexture2D labeltexture;
    QTechnique *vertexGL3Technique;
    QTechnique *vertexGL2Technique;
    QTechnique *vertexES2Technique;
    QRenderPass *vertexGL3RenderPass;
    QRenderPass *vertexGL2RenderPass;
    QRenderPass *vertexES2RenderPass;
    QShaderProgram *vertexGL3Shader;
    QShaderProgram *vertexGL2ES2Shader;
    QFilterKey *filterKey;
    QParameter param_labeltex;
};

struct Cubie {
    QEntity entity;
    QAttribute indexattribute;
    Qt3DCore::QTransform transform;
    int transformation_num;
};

struct GlobalVars {
    QEntity *puzzle_entity;
    PerVertexColorMaterial material;
    QImage atlas;
    QAttribute positionAttribute;
    QAttribute colorAttribute;
    QAttribute normalAttribute;
    QAttribute texCoordAttribute;
    QAttribute barycAttribute;
    QBuffer vertexbuffer;
    QBuffer indexbuffer;
    
    int cubie_cnt;
    QVector<Cubie*> cubies;
    
    QVector<QMatrix4x4> transformations;
    QVector<int> animation_blocks_idx;
    QVector3D animation_axis;
} *g;

void create_material(PerVertexColorMaterial* material) {
    material->material = new QMaterial();
    material->vertexEffect = new QEffect();
    material->vertexGL3Technique = new QTechnique();
    material->vertexGL2Technique = new QTechnique();
    material->vertexES2Technique = new QTechnique();
    material->vertexGL3RenderPass = new QRenderPass();
    material->vertexGL2RenderPass = new QRenderPass();
    material->vertexES2RenderPass = new QRenderPass();
    material->vertexGL3Shader = new QShaderProgram();
    material->vertexGL2ES2Shader = new QShaderProgram();
    material->filterKey = new QFilterKey;
    
    material->vertexGL3Technique->graphicsApiFilter()->setApi(QGraphicsApiFilter::OpenGL);
    material->vertexGL3Technique->graphicsApiFilter()->setMajorVersion(3);
    material->vertexGL3Technique->graphicsApiFilter()->setMinorVersion(1);
    material->vertexGL3Technique->graphicsApiFilter()->setProfile(QGraphicsApiFilter::CoreProfile);

    material->vertexGL2Technique->graphicsApiFilter()->setApi(QGraphicsApiFilter::OpenGL);
    material->vertexGL2Technique->graphicsApiFilter()->setMajorVersion(2);
    material->vertexGL2Technique->graphicsApiFilter()->setMinorVersion(0);
    material->vertexGL2Technique->graphicsApiFilter()->setProfile(QGraphicsApiFilter::NoProfile);

    material->vertexES2Technique->graphicsApiFilter()->setApi(QGraphicsApiFilter::OpenGLES);
    material->vertexES2Technique->graphicsApiFilter()->setMajorVersion(2);
    material->vertexES2Technique->graphicsApiFilter()->setMinorVersion(0);
    material->vertexES2Technique->graphicsApiFilter()->setProfile(QGraphicsApiFilter::NoProfile);

    material->filterKey->setParent(material->material);
    material->filterKey->setName(QStringLiteral("renderingStyle"));
    material->filterKey->setValue(QStringLiteral("forward"));

    material->vertexGL3Technique->addFilterKey(material->filterKey);
    material->vertexGL2Technique->addFilterKey(material->filterKey);
    material->vertexES2Technique->addFilterKey(material->filterKey);
    
    material->vertexGL3RenderPass->setShaderProgram(material->vertexGL3Shader);
    material->vertexGL2RenderPass->setShaderProgram(material->vertexGL2ES2Shader);
    material->vertexES2RenderPass->setShaderProgram(material->vertexGL2ES2Shader);
    
    material->vertexGL3Technique->addRenderPass(material->vertexGL3RenderPass);
    material->vertexGL2Technique->addRenderPass(material->vertexGL2RenderPass);
    material->vertexES2Technique->addRenderPass(material->vertexES2RenderPass);
    
    if (debug&DEBUG_SHADER_MASK) {
        if (debug&DEBUG_SHADERGL3) {
            qInfo() << "add technique GL3";
            material->vertexEffect->addTechnique(material->vertexGL3Technique);
        }
        if (debug&DEBUG_SHADERGL2) {
            qInfo() << "add technique GL2";
            material->vertexEffect->addTechnique(material->vertexGL2Technique);
        }
        if (debug&DEBUG_SHADERES2) {
            qInfo() << "add technique ES2";
            material->vertexEffect->addTechnique(material->vertexES2Technique);
        }
    } else {
        material->vertexEffect->addTechnique(material->vertexGL3Technique);
        material->vertexEffect->addTechnique(material->vertexGL2Technique);
        material->vertexEffect->addTechnique(material->vertexES2Technique);
    }
    material->material->setEffect(material->vertexEffect);
    
    material->labeltextureimage.setParent(&material->labeltexture);
    material->labeltexture.addTextureImage(&material->labeltextureimage);
    material->labeltexture.setMagnificationFilter(QAbstractTexture::Linear);
    material->labeltexture.setMinificationFilter(QAbstractTexture::Linear);
    material->labeltexture.setParent(material->material);
    material->param_labeltex.setName(QStringLiteral("tex"));
    material->param_labeltex.setValue(QVariant::fromValue(&material->labeltexture));
    material->material->addParameter(&material->param_labeltex);
}

void create_attributes(int vertexbuffer_stride) {
    int stride = 0;
    
    g->positionAttribute.setName(QAttribute::defaultPositionAttributeName());
    g->positionAttribute.setVertexBaseType(QAttribute::Float);
    g->positionAttribute.setVertexSize(3);
    g->positionAttribute.setAttributeType(QAttribute::VertexAttribute);
    g->positionAttribute.setBuffer(&g->vertexbuffer);
    g->positionAttribute.setByteStride(vertexbuffer_stride);
    g->positionAttribute.setCount(0);
    stride += 3 * sizeof(float);
    
    g->colorAttribute.setName(QAttribute::defaultColorAttributeName());
    g->colorAttribute.setVertexBaseType(QAttribute::Float);
    g->colorAttribute.setVertexSize(3);
    g->colorAttribute.setAttributeType(QAttribute::VertexAttribute);
    g->colorAttribute.setBuffer(&g->vertexbuffer);
    g->colorAttribute.setByteOffset(stride);
    g->colorAttribute.setByteStride(vertexbuffer_stride);
    g->colorAttribute.setCount(0);
    stride += 3 * sizeof(float);
    
    g->texCoordAttribute.setName(QAttribute::defaultTextureCoordinateAttributeName());
    g->texCoordAttribute.setVertexBaseType(QAttribute::Float);
    g->texCoordAttribute.setVertexSize(2);
    g->texCoordAttribute.setAttributeType(QAttribute::VertexAttribute);
    g->texCoordAttribute.setBuffer(&g->vertexbuffer);
    g->texCoordAttribute.setByteOffset(stride);
    g->texCoordAttribute.setByteStride(vertexbuffer_stride);
    g->texCoordAttribute.setCount(0);
    stride += 2 * sizeof(float);
    
    g->normalAttribute.setName(QAttribute::defaultNormalAttributeName());
    g->normalAttribute.setVertexBaseType(QAttribute::Float);
    g->normalAttribute.setVertexSize(3);
    g->normalAttribute.setAttributeType(QAttribute::VertexAttribute);
    g->normalAttribute.setBuffer(&g->vertexbuffer);
    g->normalAttribute.setByteOffset(stride);
    g->normalAttribute.setByteStride(vertexbuffer_stride);
    g->normalAttribute.setCount(0);
    stride += 3 * sizeof(float);
    
    g->barycAttribute.setName(QStringLiteral("vertexBarycentric"));
    g->barycAttribute.setVertexBaseType(QAttribute::Float);
    g->barycAttribute.setVertexSize(3);
    g->barycAttribute.setAttributeType(QAttribute::VertexAttribute);
    g->barycAttribute.setBuffer(&g->vertexbuffer);
    g->barycAttribute.setByteOffset(stride);
    g->barycAttribute.setByteStride(vertexbuffer_stride);
    g->barycAttribute.setCount(0);
    stride += 3 * sizeof(float);
    
    if (stride != vertexbuffer_stride)
        qFatal("internal error: wrong stride");
}

//pxm-FUNC _HD nogil
void init_qt_impl(int vertexbuffer_stride) {
    g = new GlobalVars();
    g->cubie_cnt = 0;
    create_material(&g->material);
    create_attributes(vertexbuffer_stride);
}

void LabelTextureImage::paint(QPainter *painter) {
    painter->setCompositionMode(QPainter::CompositionMode_Source);
    painter->drawImage(0, 0, g->atlas);
}

//pxm-FUNC _HD nogil
void set_shaders_impl(const QString &shgl3vert, const QString &shgl3frag, const QString &shes2vert, const QString &shes2frag) {
    g->material.vertexGL3Shader->setVertexShaderCode(QShaderProgram::loadSource(QUrl::fromLocalFile(shgl3vert)));
    g->material.vertexGL3Shader->setFragmentShaderCode(QShaderProgram::loadSource(QUrl::fromLocalFile(shgl3frag)));
    g->material.vertexGL2ES2Shader->setVertexShaderCode(QShaderProgram::loadSource(QUrl::fromLocalFile(shes2vert)));
    g->material.vertexGL2ES2Shader->setFragmentShaderCode(QShaderProgram::loadSource(QUrl::fromLocalFile(shes2frag)));
}

//pxm-FUNC _HD nogil
void set_atlas_image_size(int width, int height) {
    g->atlas = QImage(width, height, QImage::Format_RGBA8888);
    g->atlas.fill(0);
}

//pxm-FUNC _HD nogil
void set_atlas_image(int i, const QImage &image, int x) {
    QPainter p(&g->atlas);
    p.setCompositionMode(QPainter::CompositionMode_Source);
    p.drawImage(x, 0, image);
}

//pxm-FUNC _HD nogil
void set_atlas_texture_impl() {
    g->material.labeltextureimage.setSize(g->atlas.size());
    g->material.labeltextureimage.update();
}

//pxm-FUNC _HD nogil
void set_puzzle_entity_impl(QObject* entity) {
    g->puzzle_entity = (QEntity *)entity;
}

void create_entity(int idx) {
    Cubie* cubie = new Cubie();
    g->cubies[idx] = cubie;
    
    cubie->indexattribute.setAttributeType(QAttribute::IndexAttribute);
    cubie->indexattribute.setVertexBaseType(QAttribute::UnsignedShort);
    cubie->indexattribute.setBuffer(&g->indexbuffer);
    cubie->indexattribute.setByteOffset(0);
    cubie->indexattribute.setCount(0);
    
    QGeometry* cubie_geometry = new QGeometry();
    cubie_geometry->addAttribute(&g->positionAttribute);
    cubie_geometry->addAttribute(&g->colorAttribute);
    cubie_geometry->addAttribute(&g->normalAttribute);
    cubie_geometry->addAttribute(&g->texCoordAttribute);
    cubie_geometry->addAttribute(&g->barycAttribute);
    cubie_geometry->addAttribute(&cubie->indexattribute);
    
    QGeometryRenderer* cubie_mesh = new QGeometryRenderer(&cubie->entity);
    cubie_mesh->setGeometry(cubie_geometry);
    
    cubie->entity.addComponent(cubie_mesh);
    cubie->entity.addComponent(g->material.material);
    cubie->entity.addComponent(&cubie->transform);
}

//pxm-FUNC _HD nogil
void set_entity_count(const QByteArray &vertexbuffer, const QByteArray &indexbuffer, int nverts, int cnt) {
    g->positionAttribute.setCount(nverts);
    g->colorAttribute.setCount(nverts);
    g->texCoordAttribute.setCount(nverts);
    g->normalAttribute.setCount(nverts);
    g->barycAttribute.setCount(nverts);
    
    g->vertexbuffer.setData(vertexbuffer);
    g->indexbuffer.setData(indexbuffer);
    
    int n, old_size;
    if (cnt > g->cubies.size()) {
        old_size = g->cubies.size();
        g->cubies.resize(cnt);
        for (n = old_size; n < cnt; n++)
            create_entity(n);
    }
    if (!g->puzzle_entity) qWarning() << "g->puzzle_entity not set";
    for (n = g->cubie_cnt; n < cnt; n++)
        g->cubies[n]->entity.setParent(g->puzzle_entity);
    for (n = cnt; n < g->cubie_cnt; n++)
        g->cubies[n]->entity.setParent((QNode*)0);
    g->cubie_cnt = cnt;
}

//pxm-FUNC _HD nogil
void set_entity(int idx, int offs, int cnt) {
    auto cubie = g->cubies[idx];
    cubie->indexattribute.setByteOffset(offs);
    cubie->indexattribute.setCount(cnt);
    cubie->transform.setRotation(QQuaternion());
}

//pxm-FUNC _HD nogil
void set_entity_matrix(int idx, int b) {
    //TODO: see Qt doc for matrix property: use only rotation component
    g->cubies[idx]->transformation_num = b;
    g->cubies[idx]->transform.setMatrix(g->transformations[b]);
}

//pxm-FUNC _HD nogil
void set_animation_angle(float angle) {
    QMatrix4x4 mat;
    mat.rotate(angle, g->animation_axis);
    int idx;
    for (int i=0; i<g->animation_blocks_idx.size(); i++) {
        idx = g->animation_blocks_idx[i];
        g->cubies[idx]->transform.setMatrix(mat * g->transformations[g->cubies[idx]->transformation_num]);
    }
}

//pxm-FUNC _HD nogil
void set_animation_axis(int cnt_blocks, float axisx, float axisy, float axisz) {
    g->animation_blocks_idx.resize(cnt_blocks);
    g->animation_axis = QVector3D(axisx, axisy, axisz);
}

//pxm-FUNC _HD nogil
void set_animation_block(int _i, int i) {
    g->animation_blocks_idx[_i] = i;
}

//pxm-FUNC _HD nogil
void set_transformation_count(int cnt) {
    g->transformations.resize(cnt);
}

//pxm-FUNC _HD nogil
void set_transformation(int b, const QMatrix4x4 &mat) {
    g->transformations[b] = mat;
}

