//  Copyright © 2017-2019  B. Clausius <barcc@gmx.de>
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

// some code in class ModelSelectionItemDelegate is derived from class QCommonStyle,
// part of the QtWidgets module of the Qt Toolkit (v5.6.1):
// Original Copyright: (C) 2015 The Qt Company Ltd.
// Original license: General Public License version 2.1 or version 3
//                   as published by the Free Software Foundation

#include "qtui_impl_moc.h"
//pxd>from qt cimport *
//pxd>cdef extern from "qtui_impl_moc.h":#
//pxm>DINDENT

//pxh:
#ifndef QTUI_P_H
#define QTUI_P_H

#include <QtCore/QPropertyAnimation>
#include <QtCore/QString>
#include <QtGui/QDesktopServices>
#include <QtGui/QIconEngine>
#include <QtGui/QMouseEvent>
#include <QtGui/QPainter>
#include <QtGui/QStandardItemModel>
#include <QtGui/QTextLayout>
#include <QtWidgets/QColorDialog>
#include <QtWidgets/QLabel>
#include <QtWidgets/QLineEdit>
#include <QtWidgets/QListWidget>
#include <QtWidgets/QListWidgetItem>
#include <QtWidgets/QMenu>
#include <QtWidgets/QScrollBar>
#include <QtWidgets/QStyledItemDelegate>
#include <QtWidgets/QTreeView>
//px.

//pxh>QString gettext_translate(const char *text, const char *disambiguation);
//pxd>cdef QString gettext_translate "gettext_translate"(const char *text, const char *disambiguation) with gil:

//pxh:
static QSizeF viewItemTextLayout(QTextLayout &textLayout, int lineWidth)
{
    qreal height = 0;
    qreal widthUsed = 0;
    textLayout.beginLayout();
    while (true) {
        QTextLine line = textLayout.createLine();
        if (!line.isValid())
            break;
        line.setLineWidth(lineWidth);
        line.setPosition(QPointF(0, height));
        height += line.height();
        widthUsed = qMax(widthUsed, line.naturalTextWidth());
    }
    textLayout.endLayout();
    return QSizeF(widthUsed, height);
}

class ModelSelectionItemDelegate : public QStyledItemDelegate
{
    public:
        ModelSelectionItemDelegate() : QStyledItemDelegate() {}
        //~ModelSelectionItemDelegate();
        
        QSize sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const {
            return QSize(128, 128);
        }
        void _draw_text_background(QPainter *p, const QStyleOptionViewItem *option, const QRect &rect) const {
            QPalette::ColorGroup cg = option->widget->isEnabled()
                                      ? QPalette::Normal : QPalette::Disabled;
            if (cg == QPalette::Normal && !(option->state & QStyle::State_Active))
                cg = QPalette::Inactive;
                
            QPalette::ColorRole crole = option->state & QStyle::State_Selected ? QPalette::Highlight : QPalette::Dark;
            QColor color(option->palette.brush(cg, crole).color());
            color = color.darker(180);
            color.setAlpha(160);
            p->fillRect(rect, color);
        }
        
        void _draw_text(QPainter *p, const QStyleOptionViewItem *option, const QRect &rect) const {
            const int textMargin = option->widget->style()->pixelMetric(QStyle::PM_FocusFrameHMargin, 0, option->widget) + 1;
            
            QRect textRect = rect.adjusted(textMargin, 0, -textMargin, 0); // remove width padding
            const bool wrapText = option->features & QStyleOptionViewItem::WrapText;
            QTextOption textOption;
            textOption.setWrapMode(wrapText ? QTextOption::WordWrap : QTextOption::ManualWrap);
            textOption.setTextDirection(option->direction);
            textOption.setAlignment(QStyle::visualAlignment(option->direction, option->displayAlignment));
            QTextLayout textLayout(option->text, option->font);
            textLayout.setTextOption(textOption);
            
            QSizeF minsize(viewItemTextLayout(textLayout, textRect.width()));
            
            QString elidedText;
            qreal height = 0;
            qreal width = 0;
            int elidedIndex = -1;
            const int lineCount = textLayout.lineCount();
            for (int j = 0; j < lineCount; ++j) {
                const QTextLine line = textLayout.lineAt(j);
                if (j + 1 <= lineCount - 1) {
                    const QTextLine nextLine = textLayout.lineAt(j + 1);
                    if ((nextLine.y() + nextLine.height()) > textRect.height()) {
                        int start = line.textStart();
                        int length = line.textLength() + nextLine.textLength();
                        const QFontMetrics fontmetrics(option->font);
                        elidedText = fontmetrics.elidedText(textLayout.text().mid(start, length), option->textElideMode, textRect.width());
                        height += line.height();
                        width = textRect.width();
                        elidedIndex = j;
                        break;
                    }
                }
                if (line.naturalTextWidth() > textRect.width()) {
                    int start = line.textStart();
                    int length = line.textLength();
                    const QFontMetrics fontmetrics(option->font);
                    elidedText = fontmetrics.elidedText(textLayout.text().mid(start, length), option->textElideMode, textRect.width());
                    height += line.height();
                    width = textRect.width();
                    elidedIndex = j;
                    break;
                }
                width = qMax<qreal>(width, line.width());
                height += line.height();
            }
            
            const QRect layoutRect = QStyle::alignedRect(option->direction, option->displayAlignment,
                                                         QSize(int(width), int(height)), textRect);
            _draw_text_background(p, option, QRect(layoutRect.x()+layoutRect.width()/2-minsize.width()/2-2, layoutRect.y(),
                                                    minsize.width()+5, layoutRect.height()+1));
            const QPointF position = layoutRect.topLeft();
            for (int i = 0; i < lineCount; ++i) {
                const QTextLine line = textLayout.lineAt(i);
                if (i == elidedIndex) {
                    qreal x = position.x() + line.x();
                    qreal y = position.y() + line.y() + line.ascent();
                    p->save();
                    p->setFont(option->font);
                    p->drawText(QPointF(x, y), elidedText);
                    p->restore();
                    break;
                }
                line.draw(p, position);
            }
        }
        
        
        void paint(QPainter *p, const QStyleOptionViewItem &opt, const QModelIndex &index) const {
            QStyleOptionViewItem option = opt;
            initStyleOption(&option, index);
            
            p->save();
            p->setClipRect(option.rect);
            
            QRect iconRect = option.rect.adjusted(3, 4, -3, -6);
            QRect textRect = option.rect.adjusted(5, 6, -5, -8);
            
            // draw the background
            option.widget->style()->drawPrimitive(QStyle::PE_PanelItemViewItem, &option, p, option.widget);
            
            // draw the icon
            QIcon::Mode mode = QIcon::Normal;
            if (!(option.state & QStyle::State_Enabled))
                mode = QIcon::Disabled;
            else if (option.state & QStyle::State_Selected)
                mode = QIcon::Selected;
            QIcon::State state = option.state & QStyle::State_Open ? QIcon::On : QIcon::Off;
            option.icon.paint(p, iconRect, option.decorationAlignment, mode, state);
            
            // draw the text
            if (!option.text.isEmpty()) {
                QPalette::ColorGroup cg = option.state & QStyle::State_Enabled
                                      ? QPalette::Normal : QPalette::Disabled;
                if (cg == QPalette::Normal && !(option.state & QStyle::State_Active))
                    cg = QPalette::Inactive;
                    
                if (option.state & QStyle::State_Selected) {
                    p->setPen(option.palette.color(cg, QPalette::HighlightedText));
                } else {
                    p->setPen(option.palette.color(cg, QPalette::BrightText));
                }
                _draw_text(p, &option, textRect);
            }
            
            // draw the focus rect
            if (option.state & QStyle::State_HasFocus) {
                QStyleOptionFocusRect o;
                o.QStyleOption::operator=(option);
                o.rect = QRect(option.rect);
                o.state |= QStyle::State_KeyboardFocusChange;
                o.state |= QStyle::State_Item;
                QPalette::ColorGroup cg = (option.state & QStyle::State_Enabled)
                              ? QPalette::Normal : QPalette::Disabled;
                o.backgroundColor = option.palette.color(cg, (option.state & QStyle::State_Selected)
                                             ? QPalette::Highlight : QPalette::Window);
                option.widget->style()->drawPrimitive(QStyle::PE_FrameFocusRect, &o, p, option.widget);
            }
            
            p->restore();
        }
};

class ModelSelectionWidget : public QListWidget
{
public:
    ModelSelectionWidget(QWidget *p) : QListWidget(p) {}
    
protected:
    void mouseDoubleClickEvent(QMouseEvent *event) {
        event->ignore();
    }
    QItemSelectionModel::SelectionFlags selectionCommand(const QModelIndex &index, const QEvent *event = Q_NULLPTR) const {
        if (dynamic_cast<const QMouseEvent*>(event)) {
            return QItemSelectionModel::NoUpdate;
        } else
            return QListWidget::selectionCommand(index, event);
    }
};


class ShortcutEditor : public QLabel {
    Q_OBJECT

public:
    QString *key;

    ShortcutEditor(QWidget *parent);
    Q_SIGNAL void key_pressed();

    void keyPressEvent(QKeyEvent *event);
};

class ShortcutDelegate : public QStyledItemDelegate {
    Q_OBJECT

public:
    ShortcutDelegate(QObject *parent) : QStyledItemDelegate(parent) {}

    QWidget *createEditor(QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index) const;
    void setEditorData(QWidget *editor, const QModelIndex &index) const {}
    void setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const;

    Q_INVOKABLE void on_editor_key_pressed();
};

class ColorIconEngine : public QIconEngine
{
public:
    QString color;

    ColorIconEngine() : QIconEngine(), color(QStringLiteral("black")) {}

    void paint(QPainter *painter, const QRect &rect, QIcon::Mode mode, QIcon::State state);
    ColorIconEngine *clone() const;
};

#include "main.h"
#include "preferences.h"
#include "help.h"
#include "about.h"
//px.

///pxd/cdef cppclass UIContainer:#
//pxm-D_CPPCLASS
class UIContainer {
//pxh:
public:
    Ui::MainWindow ui_main;
    QLabel *label_selectmodel;
    int active_plugin_group;
    Ui::DialogPreferences ui_pref;
    QStandardItemModel *liststore_movekeys;
    QStandardItemModel *liststore_faces;
    int liststore_faces_width;
    ColorIconEngine *iconengine_color;
    ColorIconEngine *iconengine_background_color;
    Ui::DialogHelp ui_help;
    Ui::AboutDialog ui_about;
    int index_tab_about;
    int index_tab_license;
    QScrollBar *scrollbar;
    QPropertyAnimation animation;
//px.
    //pxm-FUNC _HD nogil
    void setupUi_main(QMainWindow *window, const QString &appicon, const QString &icon1, const QString &icon2, const QString &icon3, const QString &key1, const QString &key2);
    //pxm-FUNC _HD nogil
    void disable_tooltips();
    //pxm-FUNC _HD nogil
    void set_shortcuts(const QString &key1, const QString &key2, const QString &key3, const QString &key4);
    //pxm-FUNC _HD nogil
    void add_widgets(QMainWindow *window, QLineEdit *edit, QWidget *drawing, QLabel *status);
    //TODO: not used:
    //pxm-FUNC _HD nogil
    void set_toolbar_visible(bool visible);
    //pxm-FUNC _HD nogil
    void set_frame_editbar_visible(bool visible);
    //pxm-FUNC _HD nogil
    void set_statusbar_visible(bool visible);
    //pxm-FUNC _HD nogil
    void hide_controls();
    //pxm-FUNC _HD nogil
    void set_visible(bool editbar_visible, bool statusbar_visible);
    //pxm-FUNC _HD nogil
    void get_splitter_sizes(int &s1, int &s2);
    //pxm-FUNC _HD nogil
    void set_splitter_sizes(int divider_pos);
    //pxm-FUNC _HD nogil
    void splitter_update_minimumsize();
    //pxm-FUNC _HD nogil
    void set_toolbar_state(int a, int b, int c, int d, int e);
    //pxm-FUNC _HD nogil
    void clear_sidepane();
    //pxm-FUNC _HD nogil
    QPushButton *create_sidepane_button(const QString &text);
    //pxm-FUNC _HD nogil
    QTreeView *create_sidepane_treeview(QStandardItemModel *treestore, int i);
    //pxm-FUNC _HD nogil
    void set_active_plugin_group(int i);
    //pxm-FUNC _HD nogil
    void set_active_plugin_group_by_obj(QObject *obj);
    //pxm-FUNC _HD nogil
    int get_plugin_group_count();
    //pxm-FUNC _HD nogil
    QModelIndex root_index(int i);
    //pxm-FUNC _HD nogil
    void hide_row(int i, bool hide);
    //pxm-FUNC _HD nogil
    void set_row_hidden(int i, int row, const QModelIndex &index, bool hide);
    //pxm-FUNC _HD nogil
    int splitter_pos();
    
    //pxm-FUNC _HD nogil
    void setupUi_pref(QDialog *window, int speed, int speed_min, int speed_max, int mirror_min, int mirror_max, QStandardItemModel *_movekeys);
    //pxm-FUNC _HD nogil
    void preferences_block_signals(bool block);
    //pxm-FUNC _HD nogil
    void combobox_add_shaderitem(const QString &name, const QString &nick);
    //pxm-FUNC _HD nogil
    void combobox_add_samplesitem(const QString &name, const QString &nick);
    //pxm-FUNC _HD nogil
    void set_prefs_graphics_values(int shader, int samples, bool mirror_faces, double mirror_distance);
    //pxm-FUNC _HD nogil
    void add_movekey_row();
    //pxm-FUNC _HD nogil
    void remove_movekey_row();
    //pxm-FUNC _HD nogil
    void add_liststore_faces_row(const QString &name, const QString &key);
    //pxm-FUNC _HD nogil
    void finalize_liststore_faces(QWidget *window);
    //pxm-FUNC _HD nogil
    QString get_liststore_faces_facekey(const QModelIndex &index);
    //pxm-FUNC _HD nogil
    void set_button_color(const QString &color);
    //pxm-FUNC _HD nogil
    QString color_dialog(QWidget *parent, const QString &color);
    //pxm-FUNC _HD nogil
    void set_imagemode(int imagemode);
    //pxm-FUNC _HD nogil
    void add_combobox_image_item(const QIcon &icon, const QString &text, const QString &filename);
    //pxm-FUNC _HD nogil
    void set_combobox_image(int index_icon, const QIcon &icon, int index);
    
    //pxm-FUNC _HD nogil
    void set_button_background_color(const QString &color);
    //pxm-FUNC _HD nogil
    void fill_page(QVariant items, const QString &title);
    //pxm-FUNC _HD nogil
    void set_page(int i);
    //pxm-FUNC _HD nogil
    QDialog *setupUi_help(QWidget *window, const QString &helptext);
    //pxm-FUNC _HD nogil
    QDialog *setupUi_about(QWidget *window);
    //pxm-FUNC _HD nogil
    void fill_header(const QString &fileName, const QString &appname, const QString &version, const QString &desc);
    //pxm-FUNC _HD nogil
    void fill_about_tab(const QString &copyr, const QString &website, const QString &translators);
    //pxm-FUNC _HD nogil
    void fill_contribute_tab(const QString &contribute);
    //pxm-FUNC _HD nogil
    void fill_license_tab(const QString &license_short, bool license_notfound, const QString &license_full);
    //pxm-FUNC _HD nogil
    void tab_widget_currentChanged(int index);
    //pxm-FUNC _HD nogil
    void update_animation(bool first);
    //pxm-FUNC _HD nogil
    void event_filter_stop_animation(QEvent *event);
    //pxm-FUNC _HD nogil
    void show_full_license(bool local, const QUrl &url);
//pxm-CPPCLASS_END
};

//pxh:
class AboutDialog : public QDialog {
    Q_OBJECT
    
    UIContainer *m_qtui;
    
public:
    AboutDialog(QWidget *parent, UIContainer *qtui)
        : QDialog(parent), m_qtui(qtui) {}
        
    void showEvent(QShowEvent *unused_event) {
        m_qtui->update_animation(true);
    }
    void resizeEvent(QResizeEvent *unused_event) {
        m_qtui->update_animation(false);
    }
    bool eventFilter(QObject *unused_watched, QEvent *event) {
        m_qtui->event_filter_stop_animation(event);
        return false;
    }
    
public slots:
    void on_text_translators_anchorClicked(const QUrl &link) {
        QDesktopServices::openUrl(link);
    }
    void on_tab_widget_currentChanged(int index) {
        m_qtui->tab_widget_currentChanged(index);
    }
    void on_text_license_short_anchorClicked(const QUrl &url) {
        m_qtui->show_full_license(url.toString() == QStringLiteral("text:FULL_LICENSE_TEXT"), url);
    }
};

#endif // QTUI_P_H
//px.


//ShortcutEditor
ShortcutEditor::ShortcutEditor(QWidget *parent)
        : QLabel(gettext_translate("Press a key …", Q_NULLPTR), parent) {
    setFocusPolicy(Qt::StrongFocus);
    setAutoFillBackground(true);
}

void ShortcutEditor::keyPressEvent(QKeyEvent *event) {
    switch (event->key()) {
        case Qt::Key_Shift: case Qt::Key_Control: case Qt::Key_Meta:
        case Qt::Key_Alt: case Qt::Key_AltGr:
        case Qt::Key_CapsLock: case Qt::Key_NumLock: case Qt::Key_ScrollLock:
            QLabel::keyPressEvent(event);
            return;
    }
    if (event->count() != 1) {
        QLabel::keyPressEvent(event);
        return;
    }
    int mod = event->modifiers() & (Qt::ShiftModifier | Qt::ControlModifier);
    QStringList keylist = QKeySequence(event->key() | int(mod)).toString().split('+');
    if ((event->modifiers() & Qt::KeypadModifier) && !keylist.contains(QStringLiteral("Num")))
        keylist.insert(keylist.size()-1, "Num");
    key = new QString(keylist.join('+'));
    key_pressed();
}


//ShortcutDelegate
QWidget *ShortcutDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index) const {
    ShortcutEditor *editor = new ShortcutEditor(parent);
    connect(editor, SIGNAL(key_pressed()), this, SLOT(on_editor_key_pressed()));
    return editor;
}

void ShortcutDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const {
    if (((ShortcutEditor*)editor)->key != NULL) {
        QVariant value(*((ShortcutEditor*)editor)->key);
        model->setData(index, value, Qt::DisplayRole);
    }
}

Q_INVOKABLE void ShortcutDelegate::on_editor_key_pressed() {
    QWidget *editor = (QWidget*)this->sender();
    this->commitData(editor);
    this->closeEditor(editor, QAbstractItemDelegate::NoHint);
}


//ColorIconEngine
void ColorIconEngine::paint(QPainter *painter, const QRect &rect, QIcon::Mode mode, QIcon::State state)
    { painter->fillRect(rect, QColor(this->color)); }
ColorIconEngine *ColorIconEngine::clone() const
    {
        ColorIconEngine *cloned = new ColorIconEngine();
        cloned->color = color;
        return cloned;
    }


//UIContainer
void UIContainer::setupUi_main(QMainWindow *window,
                               const QString &appicon, const QString &icon1, const QString &icon2, const QString &icon3,
                               const QString &key1, const QString &key2) {
    // actions that belongs to no widget
    QAction *action_jump_to_editbar = new QAction(window);
    action_jump_to_editbar->setObjectName("action_jump_to_editbar");
    action_jump_to_editbar->setShortcut(QKeySequence(key1));
    window->addAction(action_jump_to_editbar);
    QAction *action_edit_cube = new QAction(window);
    action_edit_cube->setObjectName("action_edit_cube");
    action_edit_cube->setShortcut(QKeySequence(key2));
    window->addAction(action_edit_cube);
    // Qt Designer code
    ui_main.setupUi(window);
    // ...
    ui_main.splitter->setCollapsible(0, false);
    ui_main.splitter->setStretchFactor(0, 1);
    ui_main.splitter->setStretchFactor(1, 0);
    ui_main.listwidget->setItemDelegate(new ModelSelectionItemDelegate());
    // set icons
    window->setWindowIcon(QIcon(appicon));
    ui_main.button_edit_exec->setIcon(QIcon::fromTheme("system-run"));
    ui_main.button_edit_clear->setIcon(QIcon::fromTheme("edit-clear"));
    ui_main.action_challenge->setIcon(QIcon(icon1));
    ui_main.action_new_solved->setIcon(QIcon(icon2));
    ui_main.action_quit->setIcon(QIcon::fromTheme("application-exit"));
    ui_main.action_rewind->setIcon(QIcon::fromTheme("media-seek-backward"));
    ui_main.action_previous->setIcon(QIcon::fromTheme("media-skip-backward"));
    ui_main.action_stop->setIcon(QIcon::fromTheme("media-playback-stop"));
    ui_main.action_play->setIcon(QIcon::fromTheme("media-playback-start"));
    ui_main.action_next->setIcon(QIcon::fromTheme("media-skip-forward"));
    ui_main.action_forward->setIcon(QIcon::fromTheme("media-seek-forward"));
    ui_main.action_mark_set->setIcon(QIcon::fromTheme("list-add"));
    ui_main.action_mark_remove->setIcon(QIcon::fromTheme("list-remove"));
    ui_main.action_selectmodel->setIcon(QIcon(icon3));
    ui_main.action_selectmodel_back->setIcon(QIcon::fromTheme("back"));
    ui_main.action_preferences->setIcon(QIcon::fromTheme("document-properties"));
    ui_main.action_info->setIcon(QIcon::fromTheme("help-about"));
    ui_main.action_help->setIcon(QIcon::fromTheme("help"));
    // widgets that are not created with Qt Designer: toolbar_play
    ui_main.toolbar_play->addAction(ui_main.action_selectmodel);
    ui_main.toolbar_play->addAction(ui_main.action_new_solved);
    ui_main.toolbar_play->addAction(ui_main.action_challenge);
    ui_main.toolbar_play->addSeparator();
    ui_main.toolbar_play->addAction(ui_main.action_rewind);
    ui_main.toolbar_play->addAction(ui_main.action_previous);
    ui_main.toolbar_play->addAction(ui_main.action_stop);
    ui_main.toolbar_play->addAction(ui_main.action_play);
    ui_main.toolbar_play->addAction(ui_main.action_next);
    ui_main.toolbar_play->addAction(ui_main.action_forward);
    ui_main.toolbar_play->addAction(ui_main.action_mark_set);
    ui_main.toolbar_play->addAction(ui_main.action_mark_remove);
    // …: toolbar_selectmodel
    ui_main.toolbar_selectmodel->addAction(ui_main.action_selectmodel_back);
    label_selectmodel = new QLabel();
    ui_main.toolbar_selectmodel->addWidget(label_selectmodel);
    ui_main.toolbar_selectmodel->setVisible(false);
    // …: toolbar_common
    QWidget *spacer = new QWidget();
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    ui_main.toolbar_common->addWidget(spacer);
    // …: menu button
    QToolButton *menubutton = new QToolButton(window);
    menubutton->setIcon(QIcon::fromTheme("format-justify-fill")); //XXX: hmm
    menubutton->setPopupMode(QToolButton::InstantPopup);
    QMenu *menu = new QMenu(menubutton);
    menu->addAction(ui_main.action_initial_state);
    menu->addAction(ui_main.action_reset_rotation);
    menu->addSeparator();
    menu->addAction(ui_main.action_editbar);
    menu->addAction(ui_main.action_statusbar);
    menu->addSeparator();
    menu->addAction(ui_main.action_preferences);
    menu->addAction(ui_main.action_help);
    menu->addAction(ui_main.action_info);
    menu->addSeparator();
    menu->addAction(ui_main.action_quit);
    menubutton->setMenu(menu);
    ui_main.toolbar_common->addWidget(menubutton);
    // add action to main window for shortcuts
    window->addAction(ui_main.action_challenge);
    window->addAction(ui_main.action_new_solved);
    window->addAction(ui_main.action_quit);
    window->addAction(ui_main.action_selectmodel);
    window->addAction(ui_main.action_initial_state);
    window->addAction(ui_main.action_reset_rotation);
    window->addAction(ui_main.action_preferences);
    window->addAction(ui_main.action_statusbar);
    window->addAction(ui_main.action_info);
    window->addAction(ui_main.action_rewind);
    window->addAction(ui_main.action_rewind);
    window->addAction(ui_main.action_previous);
    window->addAction(ui_main.action_previous);
    window->addAction(ui_main.action_stop);
    window->addAction(ui_main.action_stop);
    window->addAction(ui_main.action_play);
    window->addAction(ui_main.action_play);
    window->addAction(ui_main.action_next);
    window->addAction(ui_main.action_next);
    window->addAction(ui_main.action_forward);
    window->addAction(ui_main.action_forward);
    window->addAction(ui_main.action_mark_set);
    window->addAction(ui_main.action_mark_set);
    window->addAction(ui_main.action_mark_remove);
    window->addAction(ui_main.action_mark_remove);
    window->addAction(ui_main.action_editbar);
    window->addAction(ui_main.action_help);
    window->addAction(ui_main.action_help);
    // rest
    disable_tooltips();
    active_plugin_group = 0;
}
void UIContainer::disable_tooltips() {
    //pita: If a action has no tooltip, the appropriate toolbutton uses the actions text for the tooltip.
    //      To remove tooltips you have to remove the tooltips for the appropriate toolbutton,
    //      but every time you change something on the action, the button will recreate the tooltip
    ui_main.toolbar_play->widgetForAction(ui_main.action_rewind)->setToolTip(QStringLiteral(""));
    ui_main.toolbar_play->widgetForAction(ui_main.action_previous)->setToolTip(QStringLiteral(""));
    ui_main.toolbar_play->widgetForAction(ui_main.action_stop)->setToolTip(QStringLiteral(""));
    ui_main.toolbar_play->widgetForAction(ui_main.action_play)->setToolTip(QStringLiteral(""));
    ui_main.toolbar_play->widgetForAction(ui_main.action_next)->setToolTip(QStringLiteral(""));
    ui_main.toolbar_play->widgetForAction(ui_main.action_forward)->setToolTip(QStringLiteral(""));
}
void UIContainer::set_shortcuts(const QString &key1, const QString &key2, const QString &key3, const QString &key4) {
    ui_main.action_selectmodel->setShortcut(QKeySequence(key1));
    ui_main.action_initial_state->setShortcut(QKeySequence(key2));
    ui_main.action_reset_rotation->setShortcut(QKeySequence(key3));
    ui_main.action_preferences->setShortcut(QKeySequence(key4));
}
void UIContainer::add_widgets(QMainWindow *window, QLineEdit *edit, QWidget *drawing, QLabel *status) {
    edit->setObjectName("edit_moves");
    edit->setFrame(false);
    ui_main.layout_moves->insertWidget(1, edit);
    drawing->setObjectName("drawingarea");
    ui_main.verticalLayout->addWidget(drawing);
    window->statusBar()->addWidget(status, 1);
    
    window->setTabOrder(edit, drawing);
    window->setTabOrder(drawing, ui_main.box_sidepane);
    drawing->setFocus(Qt::OtherFocusReason);
    
    auto sizepolicy = QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    sizepolicy.setHorizontalStretch(0);
    sizepolicy.setVerticalStretch(0);
    drawing->setSizePolicy(sizepolicy);
}
void UIContainer::set_toolbar_visible(bool visible)
    { ui_main.toolbar_play->setVisible(visible); }
void UIContainer::set_frame_editbar_visible(bool visible)
    { ui_main.frame_editbar->setVisible(visible); }
void UIContainer::set_statusbar_visible(bool visible)
    { ui_main.statusbar->setVisible(visible); }
void UIContainer::hide_controls() {
    ui_main.toolbar_play->setVisible(false);
    ui_main.toolbar_common->setVisible(false);
    ui_main.frame_editbar->setVisible(false);
    ui_main.statusbar->setVisible(false);
    ui_main.box_sidepane->setVisible(false);
}
void UIContainer::set_visible(bool editbar_visible, bool statusbar_visible) {
    ui_main.action_editbar->setChecked(editbar_visible);
    ui_main.action_statusbar->setChecked(statusbar_visible);
    ui_main.frame_editbar->setVisible(editbar_visible);
    ui_main.statusbar->setVisible(statusbar_visible);
}
void UIContainer::get_splitter_sizes(int &s1, int &s2) {
    QList<int> list(ui_main.splitter->sizes());
    s1 = list.at(0);
    s2 = list.at(1);
}
void UIContainer::set_splitter_sizes(int divider_pos) {
    QList<int> list(ui_main.splitter->sizes());
    list[1] = list[0] + list[1] - divider_pos;
    list[0] = divider_pos;
    ui_main.splitter->setSizes(list);
}
void UIContainer::splitter_update_minimumsize() {
    int w0, w1;
    get_splitter_sizes(w0, w1);
    QSize s0 = ui_main.splitter->widget(0)->minimumSizeHint();
    QSize s1 = ui_main.splitter->widget(1)->minimumSizeHint();
    int w = s0.width() + ui_main.splitter->handleWidth();
    int h = std::max(s0.height(), s1.height());
    w += w1==0 ? 6 : s1.width();
    ui_main.splitter->setMinimumSize(w, h);
}
void UIContainer::set_toolbar_state(int a, int b, int c, int d, int e) {
    ui_main.action_rewind->setEnabled(a);
    ui_main.action_previous->setEnabled(a);
    ui_main.action_play->setEnabled(b);
    ui_main.action_next->setEnabled(b);
    ui_main.action_forward->setEnabled(b);
    ui_main.action_mark_set->setEnabled(c);
    ui_main.action_mark_remove->setEnabled(c);
    ui_main.action_stop->setVisible(d);
    ui_main.action_play->setVisible(!d);
    ui_main.action_mark_set->setVisible(e);
    ui_main.action_mark_remove->setVisible(!e);
    disable_tooltips();
}
void UIContainer::clear_sidepane() {
    QLayoutItem *child;
    while ((child = ui_main.layout_sidepane->takeAt(0)) != 0) {
        delete child;
    }
}
QPushButton *UIContainer::create_sidepane_button(const QString &text) {
    QPushButton *button = new QPushButton(ui_main.box_sidepane);
    button->setText(text);
    button->setFlat(true);
    button->setFocusPolicy(Qt::TabFocus);
    ui_main.layout_sidepane->addWidget(button);
    return button;
}
QTreeView *UIContainer::create_sidepane_treeview(QStandardItemModel *treestore, int i) {
    QTreeView *treeview = new QTreeView(ui_main.box_sidepane);
    treeview->setFrameShape(QFrame::NoFrame);
    treeview->setEditTriggers(QAbstractItemView::NoEditTriggers);
    treeview->setUniformRowHeights(true);
    treeview->setAnimated(true);
    treeview->setHeaderHidden(true);
    treeview->hide();
    treeview->setModel(treestore);
    treeview->setRootIndex(treestore->index(i, 0));
    ui_main.layout_sidepane->addWidget(treeview);
    return treeview;
}
void UIContainer::set_active_plugin_group(int i) {
    QLayoutItem *item_treeview;
    if (i == -1)  i = active_plugin_group;
    item_treeview = ui_main.layout_sidepane->itemAt(active_plugin_group*2+1);
    if (item_treeview != NULL)
        item_treeview->widget()->hide();
    item_treeview = ui_main.layout_sidepane->itemAt(i*2+1);
    if (item_treeview != NULL)
        item_treeview->widget()->show();
    active_plugin_group = i;
}
void UIContainer::set_active_plugin_group_by_obj(QObject *obj) {
    int i = ui_main.layout_sidepane->indexOf((QWidget*)obj);
    this->set_active_plugin_group(i/2);
}
int UIContainer::get_plugin_group_count()
    { return ui_main.layout_sidepane->count() / 2; }
QModelIndex UIContainer::root_index(int i)
    { return ((QTreeView*)ui_main.layout_sidepane->itemAt(i*2+1)->widget())->rootIndex(); }
void UIContainer::hide_row(int i, bool hide) {
    QWidget *button = ui_main.layout_sidepane->itemAt(i*2)->widget();
    if (hide) {
        button->hide();
        if (i == active_plugin_group)  set_active_plugin_group(0);
    } else {
        button->show();
    }
}
void UIContainer::set_row_hidden(int i, int row, const QModelIndex &index, bool hide)
    { ((QTreeView*)ui_main.layout_sidepane->itemAt(i*2+1)->widget())->setRowHidden(row, index, hide); }
int UIContainer::splitter_pos()
    { return ui_main.splitter->sizes()[0]; }
void UIContainer::setupUi_pref(QDialog *window, int speed, int speed_min, int speed_max, int mirror_min, int mirror_max, QStandardItemModel *_movekeys) {
    liststore_movekeys = _movekeys;
    liststore_movekeys->setObjectName("liststore_movekeys");
    liststore_faces = new QStandardItemModel(window);
    liststore_faces->setObjectName("liststore_faces");

    ui_pref.setupUi(window);

    ui_pref.button_animspeed_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_shader_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_antialiasing_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_mirror_faces_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_movekey_add->setIcon(QIcon::fromTheme("list-add"));
    ui_pref.button_movekey_remove->setIcon(QIcon::fromTheme("list-remove"));
    ui_pref.button_movekey_reset->setIcon(QIcon::fromTheme("document-revert"));
    ui_pref.button_color_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_image_reset->setIcon(QIcon::fromTheme("edit-clear"));
    ui_pref.button_background_color_reset->setIcon(QIcon::fromTheme("edit-clear"));

    ui_pref.buttonBox->button(QDialogButtonBox::Close)->setDefault(true);
    ui_pref.label_needs_restarted->setVisible(false);
    QPalette pal(ui_pref.label_needs_restarted->palette());
    pal.setColor(QPalette::WindowText, QColor(170, 0, 0));
    ui_pref.label_needs_restarted->setPalette(pal);

    // graphic tab
    ui_pref.slider_animspeed->setRange(speed_min, speed_max);
    ui_pref.slider_animspeed->setValue(speed);
    ui_pref.spinbox_mirror_faces->setRange(mirror_min, mirror_max);

    // keys tab
    liststore_movekeys->setColumnCount(2);
    ui_pref.listview_movekeys->setModel(liststore_movekeys);
    ShortcutDelegate *shortcut_delegate = new ShortcutDelegate(ui_pref.listview_movekeys);
    ui_pref.listview_movekeys->setItemDelegateForColumn(1, shortcut_delegate);

    // theme tab
    liststore_faces_width = 30;
    iconengine_color = new ColorIconEngine();
    ui_pref.button_color->setIcon(QIcon(iconengine_color));
    int height = ui_pref.button_color->iconSize().height();
    ui_pref.button_color->setIconSize(
            QSize(height * ui_pref.button_color->width() / ui_pref.button_color->height(), height));
    ui_pref.button_color->setText(QStringLiteral(""));
    iconengine_background_color = new ColorIconEngine();
    ui_pref.button_background_color->setIcon(QIcon(iconengine_background_color));
    height = ui_pref.button_background_color->iconSize().height();
    ui_pref.button_background_color->setIconSize(
            QSize(height * ui_pref.button_background_color->width() / ui_pref.button_background_color->height(), height));
    ui_pref.button_background_color->setText(QStringLiteral(""));
}
void UIContainer::preferences_block_signals(bool block) {
    ui_pref.combobox_shader->blockSignals(block);
    ui_pref.combobox_samples->blockSignals(block);
    liststore_movekeys->blockSignals(block);
}
void UIContainer::combobox_add_shaderitem(const QString &name, const QString &nick)
    { ui_pref.combobox_shader->addItem(name, nick); }
void UIContainer::combobox_add_samplesitem(const QString &name, const QString &nick)
    { ui_pref.combobox_samples->addItem(name, nick); }
void UIContainer::set_prefs_graphics_values(int shader, int samples, bool mirror_faces, double mirror_distance) {
    ui_pref.combobox_shader->setCurrentIndex(shader);
    ui_pref.combobox_samples->setCurrentIndex(samples);
    ui_pref.checkbox_mirror_faces->setChecked(mirror_faces);
    ui_pref.spinbox_mirror_faces->setValue(mirror_distance);
}
void UIContainer::add_movekey_row() {
    int row = ui_pref.listview_movekeys->currentIndex().row();
    liststore_movekeys->insertRow(row);
    liststore_movekeys->setItem(row, 0, new QStandardItem(""));
    liststore_movekeys->setItem(row, 1, new QStandardItem(""));
    QModelIndex index = liststore_movekeys->index(row, 0);
    ui_pref.listview_movekeys->setCurrentIndex(index);
    ui_pref.listview_movekeys->edit(index);
}
void UIContainer::remove_movekey_row() {
    int row = ui_pref.listview_movekeys->currentIndex().row();
    liststore_movekeys->takeRow(row);
}
void UIContainer::add_liststore_faces_row(const QString &name, const QString &key) {
    QStandardItem *item = new QStandardItem(name);
    item->setData(QVariant(key));
    liststore_faces->appendRow(item);
    //XXX: workaround, listview_faces should automatically set to the correct width
    QFontMetrics fm(ui_pref.listview_faces->font());
    int width = fm.boundingRect(name).width() + 20;  // 20 for the scroll bar
    if (liststore_faces_width < width)  liststore_faces_width = width;
}
void UIContainer::finalize_liststore_faces(QWidget *window) {
    ui_pref.listview_faces->setMaximumWidth(liststore_faces_width);
    ui_pref.listview_faces->setModel(liststore_faces);
    QObject::connect(ui_pref.listview_faces->selectionModel(),
                       SIGNAL(currentRowChanged(const QModelIndex &, const QModelIndex &)),
                     window,
                       SLOT(_on_listview_faces_currentRowChanged(const QModelIndex &)));
    ui_pref.listview_faces->setCurrentIndex(liststore_faces->index(0, 0));
}
QString UIContainer::get_liststore_faces_facekey(const QModelIndex &index)
    { return liststore_faces->itemFromIndex(index)->data().toString(); }
void UIContainer::set_button_color(const QString &color) {
    iconengine_color->color = color;
    ui_pref.button_color->update();
}
QString UIContainer::color_dialog(QWidget *parent, const QString &color) {
    QColorDialog dialog(parent);
    dialog.setOption(QColorDialog::DontUseNativeDialog);
    dialog.setCurrentColor(QColor(color));
    if (dialog.exec() == QDialog::Accepted)
        return dialog.currentColor().name();
    return QString();
}
void UIContainer::set_imagemode(int imagemode) {
    if (imagemode == 0)
        ui_pref.radiobutton_tiled->setChecked(true);
    else if (imagemode == 1)
        ui_pref.radiobutton_mosaic->setChecked(true);
}
void UIContainer::add_combobox_image_item(const QIcon &icon, const QString &text, const QString &filename)
    { ui_pref.combobox_image->addItem(icon, text, filename); }
void UIContainer::set_combobox_image(int index_icon, const QIcon &icon, int index) {
    ui_pref.combobox_image->setItemIcon(index_icon, icon);
    if (index >= 0)  ui_pref.combobox_image->setCurrentIndex(index);
}
void UIContainer::set_button_background_color(const QString &color) {
    iconengine_background_color->color = color;
    ui_pref.button_background_color->update();
}
void UIContainer::fill_page(QVariant items, const QString &title) {
    ui_main.listwidget->clear();
    QList<QObject*> list = items.value<QList<QObject*> >();
    for (int i=0; i < list.size(); ++i) {
        QObject *obj = list.at(i);
        QString text = obj->property("text").value<QString>();
        QString filename = obj->property("key").value<QString>();
        QListWidgetItem *item = new QListWidgetItem(QIcon(filename), text);
        item->setTextAlignment(Qt::AlignHCenter | Qt::AlignBottom);
        ui_main.listwidget->addItem(item);
    }
    label_selectmodel->setText(title);
}
void UIContainer::set_page(int i) {
    ui_main.stackedwidget->setCurrentIndex(i);
    ui_main.toolbar_play->setVisible(i==0);
    ui_main.toolbar_selectmodel->setVisible(i!=0);
}
QDialog *UIContainer::setupUi_help(QWidget *window, const QString &helptext) {
    auto help = new QDialog(window);
    ui_help.setupUi(help);
    ui_help.text_help->setHtml(helptext);
    return help;
}
QDialog *UIContainer::setupUi_about(QWidget *window) {
    auto about = new AboutDialog(window, this);
    ui_about.setupUi(about);
    index_tab_about = ui_about.tab_widget->indexOf(ui_about.tab_about);
    index_tab_license = ui_about.tab_widget->indexOf(ui_about.tab_license);
    // About tab animation
    scrollbar = ui_about.text_translators->verticalScrollBar();
    animation.setTargetObject(scrollbar);
    animation.setPropertyName(QByteArray("value"));
    animation.setLoopCount(-1);
    ui_about.text_translators->viewport()->installEventFilter(about);
    return about;
}
void UIContainer::fill_header(const QString &fileName, const QString &appname, const QString &version, const QString &desc) {
    ui_about.label_icon->setPixmap(QPixmap(fileName));
    ui_about.label_appname->setText(appname);
    ui_about.label_version->setText(version);
    ui_about.label_description->setText(desc);
}
void UIContainer::fill_about_tab(const QString &copyr, const QString &website, const QString &translators) {
    ui_about.label_copyright->setText(copyr);
    ui_about.label_website->setText(website);
    ui_about.text_translators->setHtml(translators);
}
void UIContainer::fill_contribute_tab(const QString &contribute)
    { ui_about.label_contribute->setText(contribute); }
void UIContainer::fill_license_tab(const QString &license_short, bool license_notfound, const QString &license_full) {
    ui_about.text_license_short->hide();
    ui_about.text_license_full->hide();
    ui_about.text_license_short->setHtml(license_short);
    if (license_notfound)  ui_about.text_license_full->setLineWrapMode(QTextEdit::WidgetWidth);
    ui_about.text_license_full->setHtml(license_full);
}
void UIContainer::tab_widget_currentChanged(int index) {
    if (index == index_tab_about) {
        animation.resume();
        scrollbar->hide();
    } else {
        animation.pause();
        scrollbar->show();
    }
    if (index == index_tab_license) {
        ui_about.text_license_short->setVisible(true);
        ui_about.text_license_full->setVisible(false);
    }
}
void UIContainer::update_animation(bool first) {
    int smin, smax;
    if (first and animation.state() != QAbstractAnimation::Stopped) return;
    smin = scrollbar->minimum();
    smax = scrollbar->maximum();
    if (smax > smin) {
        animation.setDuration((smax-smin) * 40);
        animation.setKeyValueAt(0., smin);
        animation.setKeyValueAt(0.04, smin);
        animation.setKeyValueAt(0.50, smax);
        animation.setKeyValueAt(0.54, smax);
        animation.setKeyValueAt(1., smin);
    }
    if (first) {
        scrollbar->hide();
        animation.start();
    }
}
void UIContainer::event_filter_stop_animation(QEvent *event) {
    //assert watched == ui_about.text_translators.viewport()
    if ((event->type() == QEvent::MouseButtonPress) || (event->type() == QEvent::Wheel)) {
        animation.pause();
        scrollbar->show();
    }
}
void UIContainer::show_full_license(bool local, const QUrl &url) {
    if (local) {
        ui_about.text_license_short->setVisible(false);
        ui_about.text_license_full->setVisible(true);
    } else
        QDesktopServices::openUrl(url);
}


