#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import io
import re
import pickle
from array import array
from enum import Enum

from .debug import (debug, DEBUG_DRAW, DEBUG_DRAWNORMALS, DEBUG_DRAWAXES,
                           DEBUG_NOPICK, DEBUG_NOBEVEL, DEBUG_NOLABEL, DEBUG_NOBLACK,
                           DEBUG_NOCACHE, DEBUG_MODELFAST, DEBUG_CACHEUPDATE)
from .config import MODELS_DIR
from .utils import filebyteorder, epsilon, get_texcoords_range, Vector
from .ext import qt

try:
    _
except NameError:
    _ = lambda t: t
    

def get_texcoords(vector, normal, minx, maxx, miny, maxy):
    x, y = vector.rotationyx_normal(normal)
    return (x - minx) / (maxx - minx), (y - miny) / (maxy - miny)
    
        
class EmptyModel:
    type = ''
    symbols = ''
    symbolsI = ''
    faces = ''
    default_rotation = (0., 0.)
    
    size = sizes = ()
    blocks = []
    cell_count = 0
    cells_visible_faces = []
    
    def __init__(self, *unused_args):
        pass
        
    def __str__(self):
        return 'no puzzle'
        
        
class Model:
    bevel_width = 0.01 if DEBUG_NOBEVEL else 0.1
    
    cache_index = {}
    cache_data = {}
    cache_vectors = {} if DEBUG_CACHEUPDATE else []
    
    def __init__(self, mtype, size, noload=False):
        items = self.cache_index['type'][mtype].items()
        self.type = mtype
        for key, value in items:
            setattr(self, key, value)
        self.mformat = _(self.mformat)
        # size for UI and config, sizes for plugins
        self.size = self.cache_index['normsize'][mtype].get(size, self.defaultsize)
        self.sizes, filename = self.cache_index['sizes'][mtype][self.size]
        self.gray = {}
        
        if noload:  # tests, that need only the type and sizes attributes save a lot of time
            return
            
        self.load_data(filename)
        
        self.cell_count = len(self.cell_indices)
        self.rotation_indices = {s:i for i, s in enumerate(self.rotation_symbols)}
        self._inplace_rotations = None
        if self.slicesmode == 'all':
            self.ignored_slice = [None for sz in self.sizes]
        elif self.slicesmode == 'mid':
            self.ignored_slice = [(sz-1)//2 for sz in self.sizes]
        elif self.slicesmode == 'last1':
            self.ignored_slice = [(None if a == 1 else sz-1) for a, sz in enumerate(self.sizes)]
        else:
            assert False, self.slicesmode
        
        #XXX: self.beveled_fvertdata generated here from a nested list, but
        #     the idices come from the modeldata files
        self.beveled_fvertdata = self.gl_bevel_faces()
        
    @property
    def inplace_rotations(self):
        if self._inplace_rotations is None:
            # slow and only used in edit mode
            self._inplace_rotations = []
            for cellidx in range(self.cell_count):
                inplace_rotations = []
                for rotsym, blocknum_blocknum2 in self.rotated_position.items():
                    for blocknum, blocknum2 in enumerate(blocknum_blocknum2):
                        if blocknum == blocknum2 == cellidx and rotsym:
                            inplace_rotations.append(rotsym)
                self._inplace_rotations.append(sorted(inplace_rotations, key=len)[:2])
        return self._inplace_rotations
        
    def is_user_slice(self, iaxis, mslice):
        return self.ignored_slice[iaxis] is None or self.ignored_slice[iaxis] != mslice
        
    def __str__(self):
        return self.mformat.format(*self.size)
        
    @classmethod
    def load_index(cls):
        if DEBUG_NOCACHE:
            try:
                from buildlib import modeldata
            except ImportError:
                pass
            else:
                debug('generating model index')
                cls.cache_index = modeldata.pool_create_indexdata()
                return
        indexfilename = os.path.join(MODELS_DIR, 'f00index' if DEBUG_MODELFAST else 'd00index')
        debug('loading model index:', indexfilename)
        cls.cache_index = pickle.loads(qt.fs_read_file(indexfilename))
            
    @staticmethod
    def read_vectors(datafile):
        vlen = array('I')
        vlen.fromfile(datafile, 1)
        if sys.byteorder != filebyteorder:
            vlen.byteswap()
        vlen = vlen[0]
        vectors = array('f')
        vectors.fromfile(datafile, vlen)
        if sys.byteorder != filebyteorder:
            vectors.byteswap()
        return vectors
        
    @staticmethod
    def isplit3_vectors(vals):
        it = iter(vals)
        for x in it:
            try:
                yield Vector((x, next(it), next(it)))
            except StopIteration:
                raise ValueError('len must be a multiple of 3')
                
    @classmethod
    def _load_data(self, filename):
        if DEBUG_NOCACHE:
            try:
                from buildlib import modeldata
            except ImportError:
                pass
            else:
                print('generating model data:', filename)
                _data, _vectors, *unused = modeldata.pool_create_modelfiledata(filename)
                _vectors = [Vector(v) for v in _vectors]
                return _data, _vectors
        filename = os.path.join(MODELS_DIR, filename)
        debug('loading model data:', filename)
        datafile = io.BytesIO(qt.fs_read_file(filename))
        data = pickle.load(datafile)
        vectors = list(self.isplit3_vectors(self.read_vectors(datafile)))
        return data, vectors
        
    def load_data(self, filename):
        debug('requesting model data:', filename, self.type, self.sizes)
        try:
            data = (self.cache_data[filename] if DEBUG_CACHEUPDATE else self.cache_data)[self.type][self.sizes]
        except KeyError:
            data, vectors = self._load_data(filename)
            if DEBUG_CACHEUPDATE:
                self.cache_data[filename] = data
                self.cache_vectors[filename] = vectors
            else:
                self.__class__.cache_data = data
                self.__class__.cache_vectors = vectors
            data = data[self.type][self.sizes]
        self.vectors = self.cache_vectors[filename] if DEBUG_CACHEUPDATE else self.cache_vectors
        for key, value in data.items():
            setattr(self, key, value)
            
    @classmethod
    def from_string(cls, modelstr):
        if modelstr == '*':
            return '*', '*', (), None
        re_model = r'''^(\w+)
                        (?: \s+(\w+)
                            (?: \s*(\W)
                                \s*(\w+)
                                (?: \s*\3
                                    \s*(\w+)
                            )?)?
                            (?:\s+with
                                \s+(.+)
                        )?)?\s*$'''
        match = re.match(re_model, modelstr, re.X)
        if match is None:
            raise ValueError('Invalid model: ' + modelstr)
        #TODO: len(sizes) > 3
        mtype, width, height, depth, exp = match.group(1, 2, 4, 5, 6)
        if mtype not in cls.cache_index['types']:
            raise ValueError('Unknown model type %r' % mtype)
        def convert_if_int(value):
            if value is None:
                return value
            try:
                return int(value)
            except ValueError:
                return value
        sizes = tuple(convert_if_int(s) for s in (width, height, depth))
        return modelstr, mtype, sizes, exp
        
    def norm_symbol(self, sym):
        try:
            return self.normal_rotation_symbols[sym]
        except KeyError:
            new_sym = ''
            for c in sym:
                try:
                    new_sym = self.normal_rotation_symbols[new_sym+c]
                except KeyError:
                    raise ValueError('invalid symbol:', sym)
            return new_sym
            
    def blocksym_to_blockpos(self, blocksym):
        for i, faces in enumerate(self.cells_visible_faces):
            #FIXME: different blocks can have same visible faces (eg. edges for 4-Cubes)
            if sorted(faces) == sorted(list(blocksym)):
                return i
        raise ValueError('Invalid block symbol:', blocksym)
        
    def blockpos_to_blocksym(self, blockpos, rotsym):
        #TODO: use cells_visible_faces, see blocksym_to_blockpos
        def axisidx_to_sym(axis, idx):
            if idx <= self.sizes[axis] // 2:
                sym = self.symbols[axis]
            else:
                idx = self.sizes[axis]-1 - idx
                sym = self.symbolsI[axis]
            if idx == 0:
                # skip idx for corners
                return sym, self.face_symbolic_to_face_color(sym, rotsym)
            elif idx == 1 and self.sizes[axis] == 3:
                # for size == 3 there is only one edge
                return '', ''
            else:
                # symbol with index to distinguish edge, but no color because the face is not visible
                return sym + str(idx+1), '?'
        #FIXME: only for models with 3 axes, function only used with debug flags
        x, y, z = self.cell_indices[blockpos]
        symx, colorsymx = axisidx_to_sym(0, x)
        symy, colorsymy = axisidx_to_sym(1, y)
        symz, colorsymz = axisidx_to_sym(2, z)
        return symx + symy + symz, colorsymx + colorsymy + colorsymz
        
    def face_symbolic_to_face_color(self, face, rot):
        for k, v in self.face_permutations[rot].items():
            if v == face:
                return k
        assert False, (face, rot)
        
    def sym_to_move(self, sym, mslice=-1):
        try:
            maxis = self.symbols.index(sym)
            return maxis, mslice, False
        except ValueError:
            maxis = self.symbolsI.index(sym)
            return maxis, mslice, True
        
    def format_cubie_state(self, cell_idx, rot_sym):
        def face_slice_from_axis_slice(cells_visible_faces, a, nslice):
            face = self.symbols[a]
            if face not in cells_visible_faces:
                faceI = self.symbolsI[a]
                if faceI in cells_visible_faces:
                    face = faceI
                    nslice = self.sizes[a]-1 - nslice
                else:
                    return None, None
            return face, nslice
        position = ''
        condition = ''
        for a, nslice in enumerate(self.cell_indices[cell_idx]):
            face, nslice = face_slice_from_axis_slice(self.cells_visible_faces[cell_idx], a, nslice)
            if face is None:
                continue
            face_ = self.face_permutations[rot_sym][face].lower()
            position += face_ if nslice == 0 else face_+str(nslice+1)
            condition += face.lower() if nslice == 0 else '_'
        return f'{position}={condition}'
        
    def normalize_complete_moves(self, moves):
        syms = ''.join((self.symbols if not mdir else self.symbolsI)[maxis] for maxis, unused_mslice, mdir in moves)
        syms = self.norm_symbol(syms)
        for sym in syms:
            yield self.sym_to_move(sym)
        
    def rotate_symbolic(self, axis, rdir, block, sym):
        rsym = (self.symbols if not rdir else self.symbolsI)[axis]
        block = self.rotated_position[rsym][block]
        sym = self.norm_symbol(sym + rsym)
        return block, sym
        
    def rotate_move(self, complete_move, move):
        caxis, unused_cslice, cdir = complete_move
        maxis, mslice, mdir = move
        caxissym = (self.symbols if cdir else self.symbolsI)[caxis]
        maxissym = (self.symbols if not mdir else self.symbolsI)[maxis]
        raxissym = self.face_permutations[caxissym][maxissym]
        raxis, unused_rslice, rdir = self.sym_to_move(raxissym)
        if mdir != rdir:
            mslice = self.sizes[raxis] - 1 - mslice
        return raxis, mslice, rdir
        
    class MoveParallel (Enum):
        NO, LT, COMPAT, GT = range(4)
        
    def is_parallel(self, iaxis1, slice1, iaxis2, slice2):
        for idx in range(self.cell_count):
            pos, rot = idx, ''
            bslice = self.cell_indices[pos][iaxis1]
            if bslice != slice1:
                continue
            pos, rot = self.rotate_symbolic(iaxis1, False, pos, rot)
            
            bslice = self.cell_indices[pos][iaxis2]
            if bslice != slice2:
                continue
            pos, rot = self.rotate_symbolic(iaxis2, False, pos, rot)
            
            bslice = self.cell_indices[pos][iaxis1]
            if bslice == slice1:
                pos, rot = self.rotate_symbolic(iaxis1, True, pos, rot)
                
            bslice = self.cell_indices[pos][iaxis2]
            if bslice == slice2:
                pos, rot = self.rotate_symbolic(iaxis2, True, pos, rot)
                
            if idx != pos or rot != '':
                return False
        else:
            return True
        assert False
        
    def parallel(self, move1, move2):
        if move1.axis == move2.axis:
            if move1.slice == move2.slice:  return self.MoveParallel.COMPAT
            if move1.slice < move2.slice:   return self.MoveParallel.LT
            if move1.slice > move2.slice:   return self.MoveParallel.GT
        if self.is_parallel(move1.axis, move1.slice, move2.axis, move2.slice):
            if move1.axis < move2.axis: 
                return self.MoveParallel.LT
            else:  # move1.axis > move2.axis
                return self.MoveParallel.GT
        return self.MoveParallel.NO
        
    def _get_other_moves(self, cellidx, vector):
        '''Moves with axes that are not parallel to vector'''
        for iaxis, axis in enumerate(self.axes):
            vca = vector.cross(axis)
            if all(-epsilon < x < epsilon for x in vca):
                continue
            mslice = self.cell_indices[cellidx][iaxis]
            if mslice is not None:
                if not self.is_user_slice(iaxis, mslice):
                    mslice = -1
                yield iaxis, mslice, vca.normalised()
                
    @staticmethod
    def triangulate(indices, reverse=False):
        indices = iter(indices)
        first = next(indices)
        prev = next(indices)
        for index in indices:
            yield first
            if reverse:
                yield index
                yield prev
            else:
                yield prev
                yield index
            prev = index
            
    barycentric = {
            3:  [(1., 0., 0.),
                 (0., 1., 0.),
                 (0., 0., 1.)],
            4:  # +---+  4---3  +-1-+
                # | / |  | / |  3 2 1
                # +---+  1---2  +-3-+
                [(1., 1., 0.),
                 (0., 1., 0.),
                 (0., 1., 1.),
                 (0., 1., 0.)],
            5:  #   +---+    4---3    +---+
                #  /|  /|   /|  /|   2|  /|
                # + | / |  5 | / |  + 3 3 2
                #  \|/  |   \|/  |   1|/  |
                #   +---+    1---2    +-1-+
                [(0., 1., 1.),
                 (0., 0., 1.),
                 (1., 0., 1.),
                 (1., 0., 1.),
                 (0., 0., 1.)],
            6:  #   +---+      1---6      +---+  
                #  /|\  |\    /|\  |\    1|\  |1 
                # + | \ | +  2 | \ | 5  + 3 \ 3 +
                #  \|  \|/    \|  \|/    2|  \|2 
                #   +---+      3---4      +---+  
                [(0., 1., 1.),
                 (0., 0., 1.),
                 (1., 0., 1.),
                 (1., 0., 1.),
                 (0., 0., 1.),
                 (0., 1., 1.)],
        }
    @staticmethod
    def triangulate_barycentric(indices, reverse=False):
        count = len(indices)
        if count == 3:
            index1, index2, index3 = indices
            yield index1
            if reverse:
                yield index3; yield index2
            else:
                yield index2; yield index3
        elif count == 4:
            index1, index2, index3, index4 = indices
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
        elif count == 5:
            index1, index2, index3, index4, index5 = indices
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
                yield index1; yield index5; yield index4
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
                yield index1; yield index4; yield index5
        elif count == 6:
            index1, index2, index3, index4, index5, index6 = indices
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
                yield index1; yield index6; yield index4; yield index4; yield index6; yield index5
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
                yield index1; yield index4; yield index6; yield index4; yield index5; yield index6
        else:
            assert False, count
            
    def bevel_cell_label(self, cidx, c_faces):
        # label faces that replace the old faces
        for hf_face, hf_face_id, hf_verts in c_faces:
            fvertdata1, (fnormal, fcidx), unused = self.beveled_fvertdata[hf_face]
            if cidx != fcidx:
                fnormal = -fnormal
            yield hf_face_id, fnormal, [fvertdata1[v] for v in hf_verts]
                
    def bevel_cell_black(self, cidx, c_faces, c_edges, c_verts):
        fton = lambda fnormal, fcidx: fnormal if cidx == fcidx else -fnormal
        if not DEBUG_NOBEVEL:
            # faces that replace the old verts
            for v, vertdata in c_verts:
                vertdata = [(self.beveled_fvertdata[f][0][v], fton(*self.beveled_fvertdata[f][1])) for f in vertdata]
                yield vertdata
            # faces that replace the old edges
            for f1, f3, v1, v3 in c_edges:
                fvertdata1, finfo, unused = self.beveled_fvertdata[f1]
                normal1 = fton(*finfo)
                fvertdata3, finfo, unused = self.beveled_fvertdata[f3]
                normal3 = fton(*finfo)
                vertdata = [(fvertdata1[v3], normal1), (fvertdata3[v3], normal3),
                            (fvertdata3[v1], normal3), (fvertdata1[v1], normal1)]
                yield vertdata
        # black faces that replace the old faces
        for hf_face, hf_verts in c_faces:
            fvertdata1, finfo, unused = self.beveled_fvertdata[hf_face]
            normal1 = fton(*finfo)
            yield [(fvertdata1[v], normal1) for v in hf_verts]
            
    def gl_label_vertices(self, facesdata, faces, mirror_distance, cgray):
        if DEBUG_NOLABEL:
            return
        for faceno, normal, vertices in facesdata:
            if faceno in cgray:
                faceno = -1
            fd = faces[faceno]
            color = fd.color
            if fd.imagemode:
                texrange = self.texranges_mosaic[faceno]
            else:
                texrange = get_texcoords_range(vertices, normal)
            texpos = (get_texcoords(v, normal, *texrange) for v in vertices)
            x1,y1,x2,y2 = fd.texrect
            texpos = [(tpx*(x2-x1) + x1, tpy*(y2-y1) + y1) for tpx, tpy in texpos]
            
            vertdata = list(zip(vertices, texpos, self.barycentric[len(vertices)]))
            yield vertdata, normal, color, faceno, None
            if mirror_distance is not None:
                offset = normal * mirror_distance
                yield vertdata, -normal, color, faceno, offset
                    
    def gl_black_vertices(self, facesdata):
        if DEBUG_NOBLACK:
            return
        for vertices in facesdata:
            yield vertices
                
    def gl_bevel_faces(self):
        width = self.bevel_width
        if width == 0:
            def bevel_points(ps):
                return {p: p for p in ps}
        else:
            def bevel_points(ps):
                def bevel_point(point1, point2, point3):
                    point1 = self.vectors[point1]
                    point2 = self.vectors[point2]
                    point3 = self.vectors[point3]
                    nd1 = (point1 - point2).normalised()
                    nd2 = (point3 - point2).normalised()
                    return point2 + (nd1 + nd2) * width / nd1.cross(nd2).length()
                lps = len(ps)
                return {p: bevel_point(ps[(i-1)%lps], p, ps[(i+1)%lps]) for i, p in enumerate(ps)}
        def face_normal(p1, p2, p3, *unused_args):
            p1 = self.vectors[p1]
            p2 = self.vectors[p2]
            p3 = self.vectors[p3]
            return (p3 - p2).cross(p1 - p2).normalised()
        return [(bevel_points(ps), (face_normal(*ps), cidx), (fid, ps))
                        for cidx, cfd in enumerate(self.facesdata) for fid, ps in cfd]
                    
    def gl_vertex_data(self, theme, mirror_distance):
        vertexsize = 3+3+2+3+3
        vertexdata = []
        indexdata = []
        # list of [count vertices, face index]
        cnts_block = []
        faces = [theme.faces[fk] for fk in self.facekeys]
        faces.append(theme.gray)
        pickdata1 = {}  # keys: first vertex index; values: blockpos, symbol;
        pickdata2 = {}  # keys: blockpos, symbol; values:list of moves==(mslices, mdir);
        vidx = 0
        
        for cidx, (c_faces_label, c_faces_black, c_edges, c_verts) in enumerate(self.blocksdata):
            # bevel
            beveled_facesdata_label = self.bevel_cell_label(cidx, c_faces_label)
            beveled_facesdata_black = self.bevel_cell_black(cidx, c_faces_black, c_edges, c_verts)
            
            cnt = 0
            # label vertices
            for vertdata, n, c, faceno, offset in self.gl_label_vertices(
                                            beveled_facesdata_label, faces, mirror_distance, self.gray.get(cidx, [])):
                moves = list(self._get_other_moves(cidx, n))
                symbol = self.faces[faceno]
                for i, (v, t, b) in enumerate(vertdata, vidx):
                    if offset is not None:
                        v += offset
                    vertexdata += v
                    vertexdata += [ci/255 for ci in c]
                    vertexdata += t
                    vertexdata += n
                    vertexdata += b
                    assert i not in pickdata1
                    pickdata1[i] = (cidx, symbol)
                    #XXX: (cidx, symbol) is same for mirror faces, but moves are wrong direction
                    if (cidx, symbol) not in pickdata2:  # not a mirror face
                        pickdata2[(cidx, symbol)] = moves
                for i in self.triangulate_barycentric(list(range(vidx,vidx+len(vertdata))), reverse=(offset is not None)):
                    indexdata.append(i)
                    cnt += 1
                vidx += len(vertdata)
                assert len(vertexdata) == vidx * vertexsize, (len(vertexdata), vidx, vertexsize)
                
            # bevel, hidden labels
            for facedata in self.gl_black_vertices(beveled_facesdata_black):
                for v, n in facedata:
                    vertexdata += v
                    vertexdata += (0,0,0) #TODO: colors and texpos unused
                    vertexdata += (0,0)
                    vertexdata += n
                    vertexdata += (0,0,0)
                for i in self.triangulate(list(range(vidx,vidx+len(facedata)))):
                    indexdata.append(i)
                    cnt += 1
                vidx += len(facedata)
            cnts_block.append(cnt)
                
        nverts = len(vertexdata) // vertexsize
        assert nverts < 2**16, f'to big for indexbuffer {nverts} < {2**16}'
        vertexdata = array('f', vertexdata).tobytes()
        indexdata = array('H', indexdata).tobytes()
        debug(f'GL data: {len(vertexdata)} bytes vertexdata, {len(indexdata)} bytes indexdata')
        
        return (    self.cell_count,
                    vertexdata, vertexsize, indexdata,
                    cnts_block,
                    self.rotation_matrices,
               ), pickdata1, pickdata2
                
        
empty_model = EmptyModel()
    
    
