#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2017-2018  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
sys.path.insert(0, '.')

from contextlib import suppress
from pprint import pformat

from pybiklib import config
from pybiklib.model import Model

def format_long_line(obj):
    obj = pformat(obj, indent=0, width=100, compact=True)
    obj = '\n        '.join(obj.splitlines())
    return obj
    
def dump(model, file):
    print('Model:', model.type, model.size, file=file)
    print('Rotation-Keys:', format_long_line(sorted(model.rotated_position.keys())), file=file)
    print('Rotation-Items:', file=file)
    for sym, blocks in sorted(model.rotated_position.items()):
        print('   ', repr(sym), ':', format_long_line(blocks), file=file)
    print('Cells:', format_long_line(model.cell_indices), file=file)
    cells_visible_faces = [''.join(vf) for vf in model.cells_visible_faces]
    print('Faces:', format_long_line(cells_visible_faces), file=file)
    print(file=file)
    
def run(version):
    with open('tests/data/gamesave_interface.dump', 'wt', encoding='utf-8') as dumpfile:
        print('# generated with:', *sys.argv, file=dumpfile)
        print(file=dumpfile)
        print('Version:', version, file=dumpfile)
        print(file=dumpfile)
        try:
            mtypes = Model.cache_index['types']
            sizes = lambda mtype: sorted(set(Model.cache_index['normsize'][mtype].values()))
        except AttributeError:
            # Version 2.0
            mtypes = [t for t,u in Model.cache['index']['sizenames']]
            sizes = lambda mtype: sorted(set(i for i,f in Model.cache['index']['size'][mtype].values()))
        for mtype in mtypes:
            if version == '2.0' and mtype not in ['Cube', 'Tower', 'Brick', 'Tetrahedron']:
                continue
            if version == '2.1' and mtype in ['Octahedron', 'Dodecahedron']:
                continue
            for size in sizes(mtype):
                #if size and max(size) > 4:
                #    continue
                print(mtype, size, end=': ', flush=True)
                model = Model(mtype, size)
                print(str(model), flush=True)
                dump(model, dumpfile)
    print('done')
    
versions = ['2.0', '2.1', '3.0', config.VERSION, config.VERSION.split('~')[0]]

def usage():
    print('usage: ', os.path.basename(__file__), '[current|'+'|'.join(versions)+']')
    sys.exit(1)
    
def main():
    args = sys.argv[1:]
    if len(args) != 1:
        usage()
    version = args[0]
    if version == 'current':
        version = config.VERSION
    if version not in versions:
        usage()
    Model.load_index()
    run(version)
    
    
if __name__ == '__main__':
    main()
    

