#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2016  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
from contextlib import suppress
from threading import Thread
from queue import Queue, Empty
from subprocess import Popen, PIPE

if '-v' in sys.argv[1:]:
    info = lambda *args: print('inotify-filter', *args, file=sys.stderr)
else:
    info = lambda *args: None
    
def inotify_loop(queue):
    def events():
        for e in 'create attrib modify moved_to moved_from delete move_self delete_self'.split():
            yield '-e'
            yield e
    excludes = ['@./'+e for e in 'build click junk temp'.split()]
    pocess = Popen(['inotifywait', '-rmq', '--format', '%e %w %f', '.'] + excludes + list(events()),
                   stdout=PIPE, universal_newlines=True)
    assert os.get_blocking(pocess.stdout.fileno())
    try:
        while True:
            line = pocess.stdout.readline()
            if not line:
                pocess.wait()
                break
            queue.put(line)
    except KeyboardInterrupt:
        pocess.terminate()
        
def emit_file(filenames, msg):
    filename = filenames.pop(0)
    if os.path.isfile(filename):
        info('emit (%s):'%msg, repr(filename), len(filenames))
        print(filename, flush=True)
        sys.stdout.flush()
    else:
        info('not a file:', repr(filename), len(filenames))
    
def clean(filenames):
    while len(filenames):
        emit_file(filenames, 'clean')
        
def process_event(filenames, line):
    event, path, basename = line.split(' ', 2)
    event = event.split(',')
    assert basename and basename[-1] == '\n'
    basename = basename[:-1]
    filename = path + basename
    
    if 'ISDIR' in event:
        info('(%s) ignore directories for now:'%','.join(event), repr(filename), len(filenames))
        #XXX: files in moved directories are not handled correctly
        return
    assert len(event) == 1, event
    event = event[0]
    
    if basename.startswith('.'):
        info('(%s) ignore junk:'%event, repr(basename), len(filenames))
    elif event in ['MOVED_FROM', 'DELETE']:
        with suppress(ValueError):
            filenames.remove(filename)
        info('(%s) ignore delete for now:'%event, repr(filename), len(filenames))
    elif event in ['MOVE_SELF', 'DELETE_SELF']:
        info('(%s) ignore delete self for now:'%event, repr(filename), len(filenames))
    elif filename in filenames:
        info('(%s) ignore duplicate:'%event, repr(filename), len(filenames))
    else:
        filenames.append(filename)
        info('(%s) enqueue:'%event, repr(filename), len(filenames))
    if len(filenames) > 10:
        emit_file(filenames, 'full')
        
def emit_loop(queue):
    info('emit loop: started')
    filenames = []
    while True:
        try:
            line = queue.get(block=True, timeout=(0.1 if filenames else None))
        except Empty:
            emit_file(filenames, 'wait')
            continue
        if line is None:
            info('emit loop:', repr(line), len(filenames))
            clean(filenames)
            break
        try:
            process_event(filenames, line)
        finally:
            queue.task_done()
    info('emit loop: done')
    
def main():
    info('main: started')
    queue = Queue()
    thread = Thread(target=emit_loop, args=(queue,))
    info('main: starting thread')
    thread.start()
    info('main: starting inotify')
    inotify_loop(queue)
    info('main: inotify done')
    queue.join()
    info('main: queue empty')
    queue.put(None)
    thread.join()
    info('main: thread done')
    
if __name__ == '__main__':
    main()
    













