#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2016-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, sysconfig
from shutil import copy, rmtree
from subprocess import check_output, check_call, STDOUT
from glob import glob

sys.path.insert(0, '.')
from buildlib.create_docs import create_doc
from buildlib.utils import modify_file
from pybiklib import config


def check_output_str(cmd):
    return check_output(cmd.split()).strip().decode('utf-8')
    
def copy_files(files):
    for file in files:
        if type(file) is tuple:
            file, dstfile = file
        else:
            dstfile = file
        if not file or file[0] == '#' or not os.path.isfile(file):
            continue
        dstfile = 'click/'+dstfile
        os.makedirs(os.path.dirname(dstfile), exist_ok=True)
        copy(file, dstfile)
        
def stage_copy_indep():
    # click files
    create_doc('manifest.json', 'click/manifest.json')
    files = 'data/app-meta/pybik.desktop tools/phone/pybik.apparmor tools/phone/pybik.sh'.split()
    for file in files:
        copy(file, 'click')
    # pybik files
    with open('MANIFEST', 'rt', encoding='utf-8') as file:
        files = file.read().splitlines()
    files += glob('data/models/*')
    files += glob('data/locale/**', recursive=True)
    copy_files(files)
    # change some files
    create_doc('README', 'click/README', skip='Installation')
    os.replace('click/pybiklib/pybik.py', 'click/pybik.py')
    modify_file('click/pybik.desktop', [
            (r'^Exec=.*$','Exec=pybik.sh'),
            (r'^Icon=.*$','Icon=data/icons/pybik-192.png'),
            (r'^#?X-Ubuntu-Gettext-Domain=','#X-Ubuntu-Gettext-Domain='),
            (r'^#?X-Ubuntu-Touch=','X-Ubuntu-Touch='),
        ])
    
def find_python_imports():
    for pattern in ('pybiklib/*', 'pybiklib/*/*'):
        for filename in glob(pattern):
            if not filename.endswith('.py') or not os.path.isfile(filename):
                continue
            with open(filename, 'rt', encoding='utf-8') as fr:
                for line in fr:
                    line = line.strip()
                    if line.startswith(('from __future__',
                                        'from .', 'from pybiklib ', 'import pybiklib.', 'from pybiktest.',
                                        'from PyQt5.', 'from OpenGL.', 'import OpenGL.')
                                       ) or line in ['import pybiklib', 'import OpenGL']:
                        continue
                    if line.startswith('import ') or line.startswith('from '):
                        pos = line.find('(')
                        if pos > 0:
                            line = line[:pos] + '*'
                        yield line
                        
def find_python_stdlib_files(imports):
    import re
    lines = check_output('python3 -Bv'.split(), stderr=STDOUT, input=imports.encode('utf-8'))
    lines = lines.strip().decode('utf-8').splitlines()
    for line in lines:
        match = re.match(r"# code object from '(.*)'", line)
        if match is None:
            match = re.match(r"# extension module loaded from '(.*)'", line)
        if match:
            yield match.group(1)
            
def stage_copy_arch():
    files = []
    stdlib = sysconfig.get_path('stdlib')
    platdir = sysconfig.get_config_var('PLATDIR')
    
    # pybik files
    modify_file('click/pybik.sh', [
            (r'^#?export PYTHONHOME=.*$','export PYTHONHOME=.'+stdlib),
            (r'^#?export PYTHONPATH=.*$','export PYTHONPATH=.:.{0}:.{0}/{1}:.{0}/lib-dynload'.format(stdlib, platdir)),
        ])
    import py_compile
    for pat in 'click/pybiklib/*', 'click/pybiklib/*/*':
        for file in glob(pat):
            if file.endswith('.pxd'):
                os.remove(file)
            elif file.endswith('.py') and os.path.isfile(file):
                py_compile.compile(file, file+'c')
                os.remove(file)
    check_call('./setup.py build_ext -i --variant=es2'.split())
    files += glob('pybiklib/ext/*'+sysconfig.get_config_var('SO'))
    copy_files(files)
    
    # bundle python
    from importlib.util import cache_from_source
    imports = '\n'.join(set(find_python_imports()))+'\n'
    whitelist = set(f for f in find_python_stdlib_files(imports))
    def filter_files(files):
        for file in files:
            if not file.startswith(stdlib):
                continue
            elif file.endswith('.py'):
                cachefile = cache_from_source(file)
                if cachefile in whitelist:
                    yield cachefile, file+'c'
            elif file.endswith('.so'):
                if file in whitelist:
                    yield file
    copy('/usr/bin/python3', 'click')
    files = check_output_str('dpkg -L libpython{0.major}.{0.minor}-minimal'.format(sys.version_info)).splitlines()
    files += check_output_str('dpkg -L libpython{0.major}.{0.minor}-stdlib'.format(sys.version_info)).splitlines()
    files = filter_files(files)
    copy_files(files)
        
    # bundle qml modules
    files = check_output_str('dpkg -L qml-module-qtquick-controls').splitlines()
    files += check_output_str('dpkg -L qml-module-qtquick-dialogs').splitlines()
    files += check_output_str('dpkg -L qml-module-qtquick-privatewidgets').splitlines()
    copy_files(files)
    
def stage_build_click():
    for f in 'buildlib data/templates data/tests po pybiktest tools usr/share'.split():
        rmtree('click/'+f)
    for f in 'data/GPL-3 build_local.sh INSTALL setup.py'.split():
        os.remove('click/'+f)
    check_call('click build click --no-validate'.split())
    os.rename('{0.PACKAGE_FULL}_{0.VERSION_SHORT}_armhf.click'.format(config), 'pybik.click')
    
def main():
    args = sys.argv[1:]
    os.makedirs('click', exist_ok=True)
    if not args:
        stage_copy_indep()
        stage_copy_arch()
        stage_build_click()
    else:
        assert len(args) == 1
        funcname = 'stage_'+args[0]
        print('****', funcname, '****')
        globals()[funcname]()
        
    
if __name__ == '__main__':
    main()
    

