#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import sys, os
from glob import glob
from subprocess import run, PIPE
from collections import namedtuple, OrderedDict
from multiprocessing import Pool, cpu_count


def po_edit(po_file):
    msgid = 'msgid "wrtlbrmft"'
    msgstr_search  = 'msgstr "a'
    msgstr_replace = 'msgstr "b'
    
    print(po_file+':')
    with open(po_file, 'rt', encoding='utf-8') as f:
        lines = f.read().split('\n')
    found = False
    changed = False
    for i, line in enumerate(lines):
        if line == msgid:
            found = True
        if found and not line:
            break
        if found and msgstr_search in line:
            lines[i] = line.replace(msgstr_search, msgstr_replace)
            changed = True
            print('  Replaced in line {}'.format(po_file, i+1))
            break
    if changed:
        with open(po_file, 'wt', encoding='utf-8') as f:
            f.write('\n'.join(lines))
            
def po_mergeall(po_file):
    print((po_file+': ').ljust(filename_max_len+2), end='', flush=True)
    run(['msgmerge', '-U', po_file, 'po/pybik.pot'])
        
class POChanged:
    def __init__(self, po_file):
        self.po_file = po_file
        self.message = None
        
    def _parse_entry(self, lines):
        line = lines.pop(0) if lines else None
        comments = []
        while line and line.startswith(('# ', '#. ')):
            comments.append(line)
            line = lines.pop(0) if lines else None
        positions = []
        while line and line.startswith('#: '):
            positions.append(line)
            line = lines.pop(0) if lines else None
        flags = []
        while line and line.startswith('#'):
            flags.append(line)
            line = lines.pop(0) if lines else None
        msgid = []
        while line and line.startswith(('msgid ', '"')):
            msgid.append(line)
            line = lines.pop(0) if lines else None
        msgrest = []
        while line and line.startswith(('msg', '"')):
            msgrest.append(line)
            line = lines.pop(0) if lines else None
        assert not lines, lines
        msgid = tuple(msgid)
        return namedtuple('entry', 'comments positions flags msgid msgrest')(comments, positions, flags, msgid, msgrest)
        
    def _split_po(self, text):
        entries = []
        entry = []
        for line in text.split('\n'):
            if line:
                entry.append(line)
            elif entry:
                entries.append(self._parse_entry(entry))
                entry = []
        return entries
        
    def init_message(self):
        self.message = (self.po_file+': ').ljust(filename_max_len+2)
        
    def run(self, removed=True, added=False, comments=False, positions=False, flags=False, msg=False):
        from difflib import SequenceMatcher
        sequencematcher = SequenceMatcher(lambda s: s in '"')
        ration_pre = lambda s: ''.join(s).replace('msgid', '')
        def ratio(s1, s2):
            sequencematcher.set_seqs(ration_pre(s1), ration_pre(s2))
            return sequencematcher.ratio()
            
        old_po = run(['bzr', 'cat', self.po_file], stdout=PIPE, universal_newlines=True).stdout
        old_po = self._split_po(old_po)
        old_msgids = [e.msgid for e in old_po]
        cnt_old = len(old_po)
        
        with open(self.po_file, 'rt', encoding='utf-8') as f:
            new_po = f.read()
        new_po = self._split_po(new_po)
        new_msgids = [e.msgid for e in new_po]
        cnt_new = len(new_po)
        
        offs = 0
        
        while True:
            if old_po:
                old = old_po[0]
                old_msgid = old.msgid
                assert old_msgid == old_msgids[0]
            else:
                old = old_msgid = None
            assert offs <= len(new_po)
            if new_po and offs < len(new_po):
                new = new_po[offs]
                new_msgid = new.msgid
                assert new_msgid == new_msgids[offs]
            else:
                new = new_msgid = None
            if old_msgid and new_msgid and old_msgid == new_msgid:
                # equal msgid
                old_po.pop(0)
                old_msgids.pop(0)
                new_po.pop(offs)
                new_msgids.pop(offs)
                assert old.msgid == new.msgid == old_msgid == new_msgid
            elif old_msgid and new_msgid and new_msgid in old_msgids:
                # entry processed later, not block next entries in new_po
                assert new_msgid != old_msgid
                offs += 1
                continue
            elif old_msgid and new_msgid and old_msgid not in new_msgids and ratio(old_msgid, new_msgid)>0.5:
                # possibly changed msgid
                old_po.pop(0)
                old_msgids.pop(0)
                new_po.pop(offs)
                new_msgids.pop(offs)
                assert old.msgid != new.msgid == new_msgid
            elif old_msgid and old_msgid not in new_msgids:
                # removed entry
                self.entry_removed(old, removed)
                old_po.pop(0)
                old_msgids.pop(0)
                offs = 0
                continue
            elif new_msgid and new_msgid not in old_msgids:
                # added entry
                self.entry_added(new, added)
                new_po.pop(offs)
                new_msgids.pop(offs)
                offs = 0
                continue
            elif old_msgid:
                # entry with changed position
                new_idx = new_msgids.index(old_msgid)
                old_po.pop(0)
                old_msgids.pop(0)
                new = new_po.pop(new_idx)
                new_msgid = new_msgids.pop(new_idx)
                assert old.msgid == new.msgid == old_msgid == new_msgid
            else:
                # all entries left are obsolete
                assert all(not m for m in old_msgids)
                assert all(not m for m in new_msgids)
                break
            # revert part of entry
            assert old_msgid and new_msgid, (old_msgid, new_msgid)
            res = new
            if not comments:
                res = res._replace(comments=old.comments)
            if not positions:
                res = res._replace(positions=old.positions)
            if not flags:
                res = res._replace(flags=old.flags)
            if not msg:
                res = res._replace(msgid=old.msgid)
                res = res._replace(msgrest=old.msgrest)
            self.entry_changed(old, new, res)
            offs = 0
        # take old entries with obsolete msgid
        while old_po:
            old = old_po.pop(0)
            old_msgid = old_msgids.pop(0)
            assert not old_msgid and not old.msgid
            self.entry_obsolete(old, True)
        # discard new entries with obsolete msgid
        while new_po:
            new = new_po.pop(0)
            new_msgid = new_msgids.pop(0)
            assert not new_msgid and not new.msgid, (new_msgid, new.msgid)
            self.entry_obsolete(old, False)
            
        return cnt_old, cnt_new
        
class PORevertParts (POChanged):
    def __init__(self, po_file, **kwargs):
        super().__init__(po_file)
        self.reverted = 0
        self.changed = 0
        self.obsoleted = 0
        self.res_po = []
        
        self.init_message()
        cnt_old, cnt_new = self.run(**kwargs)
        self.write_po(cnt_old, cnt_new)
        
    def entry_removed(self, old, removed):
        if removed:
            self.res_po.append(old)
            self.reverted += 1
        else:
            self.changed += 1
        
    def entry_added(self, new, added):
        if added:
            self.res_po.append(new)
            self.changed += 1
        else:
            self.reverted += 1
        
    def entry_changed(self, old, new, res):
        if res != new:
            self.reverted += 1
        if res != old:
            self.changed += 1
        self.res_po.append(res)
        
    def entry_obsolete(self, entry, old):
        if old:
            self.res_po.append(entry)
            self.obsoleted -= 1
        else:
            self.obsoleted += 1
        
    def write_po(self, cnt_old, cnt_new):
        cnt_res = len(self.res_po)
        #FIXME: can not determine reliably whether content changes due to reordered entries
        #if self.reverted or self.obsoleted:
        if True:
            with open(self.po_file, 'wt', encoding='utf-8') as f:
                for i, entry in enumerate(self.res_po):
                    if i:
                        f.write('\n')
                    for lines in entry:
                        for line in lines:
                            f.write(line+'\n')
            self.message += '{} entries reverted, {} changed entries, (#{},#{})->#{}'.format(
                                        self.reverted, self.changed, cnt_old, cnt_new, cnt_res)
        else:
            self.message += 'nothing reverted, {} changed entries, (#{},#{})'.format(self.changed, cnt_old, cnt_new)
            assert cnt_res == cnt_new, (cnt_res, cnt_new)
        
def po_revertparts_p(po_file):
    kwargs = dict(
            #removed=False,
            #added=True,
            #comments=True,
            #positions=True,
            flags=True,
            msg=True,
        )
    return PORevertParts(po_file, **kwargs).message
    
def po_revert_keep_changed_p(po_file):
    return PORevertParts(po_file, flags=True, msg=True).message
    
def po_revert_keep_removed_p(po_file):
    return PORevertParts(po_file, removed=False).message
    
def po_revert_keep_added_p(po_file):
    return PORevertParts(po_file, added=True).message
    
def po_revert_keep_positions_p(po_file):
    return PORevertParts(po_file, positions=True).message
    
def po_revert_keep_meta_p(po_file):
    return PORevertParts(po_file, comments=True, positions=True, flags=True).message
    
class POChangedMsgids (POChanged):
    msgidstat = OrderedDict()
    
    def __init__(self, po_file, **kwargs):
        super().__init__(po_file)
        self.lang = os.path.basename(self.po_file).split('.')[0]
        
        print(self.lang, end=' ', flush=True)
        self.run()
        
    def entry_removed(self, old, removed):
        self.msgidstat.setdefault(old.msgid, {}).setdefault(0, []).append(self.lang)
        
    def entry_added(self, new, added):
        self.msgidstat.setdefault(new.msgid, {}).setdefault(1, []).append(self.lang)
        
    def entry_changed(self, old, new, res):
        if old.comments != new.comments:
            self.msgidstat.setdefault(new.msgid, {}).setdefault(2, []).append(self.lang)
        if old.positions != new.positions:
            self.msgidstat.setdefault(new.msgid, {}).setdefault(3, []).append(self.lang)
        if old.flags != new.flags:
            self.msgidstat.setdefault(new.msgid, {}).setdefault(4, []).append(self.lang)
        if old.msgid != new.msgid or old.msgrest != new.msgrest:
            self.msgidstat.setdefault(new.msgid, {}).setdefault(5, []).append(self.lang)
        
    def entry_obsolete(self, entry, old):
        pass
            
    @classmethod
    def result(self):
        print()
        modes = 'removed added comment position flags message'.split()
        for msgid, stat in self.msgidstat.items():
            for m in msgid:
                print(m)
            for mode, langs in sorted(stat.items()):
                mode = modes[mode]
                print(' ', mode+':', ' '.join(langs))
                
def po_changedmsgids(po_file=None):
    if po_file is not None:
        POChangedMsgids(po_file)
    else:
        POChangedMsgids.result()
    
def po__h(po_file=None, res=0):
    print('usage:')
    print(' ', os.path.basename(__file__), 'command [file]')
    print()
    print('command:', *sorted(commands.keys()))
    sys.exit(res)
    
commands = {}
def _commands():
    for name, func in globals().items():
        if name.startswith('po_'):
            name = name[3:]
            parallel = name.endswith('_p')
            if parallel:
                name = name[:-2]
            name = name.replace('_', '-')
            commands[name] = func, parallel
            
def main():
    _commands()
    args = sys.argv[1:]
    if len(args) == 0:
        po__h(res=0)
    elif len(args) == 1:
        files = sorted(glob('po/*.po'))
    elif len(args) == 2:
        files = [args[1]]
    else:
        po__h(res=1)
    try:
        func, parallel = commands[args[0]]
    except KeyError:
        po__h(res=1)
    global filename_max_len
    filename_max_len = max(len(f) for f in files)
    if parallel:
        pool = Pool(processes=cpu_count())
        for msg in pool.imap(func, files):
            print(msg)
    else:
        for po_file in files:
            msg = func(po_file)
            if msg is not None:
                print(msg)
        if func.__defaults__ == (None,):
            func()
    
if __name__ == '__main__':
    main()
    

